from pathlib import Path

from cg.constants import FileExtensions
from cg.constants.pacbio import PacBioDirsAndFiles
from cg.services.run_devices.abstract_classes import RunFileManager
from cg.services.run_devices.error_handler import handle_post_processing_errors
from cg.services.run_devices.exc import PostProcessingRunFileManagerError
from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData
from cg.services.run_devices.validators import validate_files_or_directories_exist
from cg.utils.files import get_files_matching_pattern


class PacBioRunFileManager(RunFileManager):

    @handle_post_processing_errors(
        to_except=(FileNotFoundError,), to_raise=PostProcessingRunFileManagerError
    )
    def get_files_to_parse(self, run_data: PacBioRunData) -> list[Path]:
        """Get the file paths required by the PacBioMetricsParser."""
        run_path: Path = run_data.full_path
        validate_files_or_directories_exist([run_path])
        return self._get_report_files(run_path)

    @handle_post_processing_errors(
        to_except=(FileNotFoundError,), to_raise=PostProcessingRunFileManagerError
    )
    def get_files_to_store(self, run_data: PacBioRunData) -> list[Path]:
        """Get the files to store for the PostProcessingHKService."""
        run_path: Path = run_data.full_path
        return self.get_files_to_parse(run_data) + self._get_hifi_read_files(run_path)

    @staticmethod
    def _get_ccs_report_file(run_path: Path) -> Path:
        """Return the path to the CCS report file."""
        statistics_dir: Path = Path(run_path, PacBioDirsAndFiles.STATISTICS_DIR)
        files: list[Path] = get_files_matching_pattern(
            directory=statistics_dir, pattern=f"*{PacBioDirsAndFiles.CCS_REPORT_SUFFIX}"
        )
        if not files:
            raise FileNotFoundError(f"No CCS report file found in {statistics_dir}")
        return files[0]

    def _get_report_files(self, run_path: Path) -> list[Path]:
        """Return the paths to the unzipped report files."""
        unzipped_dir: Path = Path(
            run_path, PacBioDirsAndFiles.STATISTICS_DIR, PacBioDirsAndFiles.UNZIPPED_REPORTS_DIR
        )
        report_files: list[Path] = [
            Path(unzipped_dir, PacBioDirsAndFiles.CONTROL_REPORT),
            Path(unzipped_dir, PacBioDirsAndFiles.LOADING_REPORT),
            Path(unzipped_dir, PacBioDirsAndFiles.RAW_DATA_REPORT),
            Path(unzipped_dir, PacBioDirsAndFiles.SMRTLINK_DATASETS_REPORT),
            self._get_ccs_report_file(run_path),
        ]
        validate_files_or_directories_exist(report_files)
        return report_files

    @staticmethod
    def _get_hifi_read_files(run_path: Path) -> list[Path]:
        """Return the path to the HiFi read file."""
        hifi_dir = Path(run_path, PacBioDirsAndFiles.HIFI_READS)
        bam_files: list[Path] = get_files_matching_pattern(
            directory=hifi_dir, pattern=f"*{FileExtensions.BAM}"
        )
        validate_files_or_directories_exist(bam_files)
        return bam_files



from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.services.deliver_files.delivery_file_tag_fetcher_service.delivery_file_tag_fetcher_service import (
    FetchDeliveryFileTagsService,
)
from cg.services.deliver_files.delivery_file_fetcher_service.analysis_delivery_file_fetcher import (
    AnalysisDeliveryFileFetcher,
)
from cg.services.deliver_files.delivery_file_fetcher_service.delivery_file_fetcher_service import (
    FetchDeliveryFilesService,
)
from cg.services.deliver_files.delivery_file_fetcher_service.fastq_delivery_file_fetcher import (
    FastqDeliveryFileFetcher,
)
from cg.services.deliver_files.delivery_file_fetcher_service.models import (
    DeliveryFiles,
    DeliveryMetaData,
)
from cg.store.models import Case
from cg.store.store import Store


class FastqAndAnalysisDeliveryFileFetcher(FetchDeliveryFilesService):

    def __init__(
        self, status_db: Store, hk_api: HousekeeperAPI, tags_fetcher: FetchDeliveryFileTagsService
    ):
        self.status_db = status_db
        self.hk_api = hk_api
        self.tags_fetcher = tags_fetcher

    def get_files_to_deliver(self, case_id: str) -> DeliveryFiles:
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        fastq_files: DeliveryFiles = self._fetch_files(
            service_class=FastqDeliveryFileFetcher, case_id=case_id
        )
        analysis_files: DeliveryFiles = self._fetch_files(
            service_class=AnalysisDeliveryFileFetcher, case_id=case_id
        )
        delivery_data = DeliveryMetaData(
            customer_internal_id=case.customer.internal_id, ticket_id=case.latest_ticket
        )

        return DeliveryFiles(
            delivery_data=delivery_data,
            case_files=analysis_files.case_files,
            sample_files=analysis_files.sample_files + fastq_files.sample_files,
        )

    def _fetch_files(self, service_class: type, case_id: str) -> DeliveryFiles:
        """Fetch files using the provided service class."""
        service = service_class(self.status_db, self.hk_api, tags_fetcher=self.tags_fetcher)
        return service.get_files_to_deliver(case_id)



"""CLI utility methods."""

import click


def validate_force_store_option(force: bool, comment: str | None):
    """Validate that the --comment option is provided when --force is set."""
    if force and not comment:
        raise click.UsageError("The --comment option is required when --force is used")



from pathlib import Path

from cg.constants import FileExtensions
from cg.meta.upload.fohm.fohm import FOHMUploadAPI
from cg.models.fohm.reports import FohmComplementaryReport, FohmPangolinReport


def test_create_daily_delivery(fohm_upload_api: FOHMUploadAPI, csv_file_path: Path):
    # GIVEN a list of CSV files

    # WHEN creating the reports content
    contents: list[dict] = fohm_upload_api.get_reports_contents([csv_file_path, csv_file_path])

    # THEN each file is a list of dicts where each dict represents a row in a CSV file
    assert isinstance(contents[0], dict)

    # THEN two files are added as a list of dicts
    assert len(contents) == 6


def test_validate_fohm_complementary_reports(
    fohm_upload_api: FOHMUploadAPI, fohm_complementary_report_raw: dict[str, str]
):
    # GIVEN a dict

    # WHEN validating the dict
    content: list[FohmComplementaryReport] = fohm_upload_api.validate_fohm_complementary_reports(
        [fohm_complementary_report_raw]
    )

    # THEN a list of reports is returned
    assert isinstance(content[0], FohmComplementaryReport)


def test_validate_fohm_pangolin_reports(
    fohm_upload_api: FOHMUploadAPI, fohm_pangolin_report_raw: dict[str, str]
):
    # GIVEN a dict

    # WHEN validating the dict
    content: list[FohmPangolinReport] = fohm_upload_api.validate_fohm_pangolin_reports(
        [fohm_pangolin_report_raw]
    )

    # THEN a list of reports is returned
    assert isinstance(content[0], FohmPangolinReport)


def test_get_sars_cov_complementary_reports(
    fohm_complementary_reports: list[FohmComplementaryReport], fohm_upload_api: FOHMUploadAPI
):
    # GIVEN a list of reports

    # WHEN getting Sars-cov reports from reports
    content: list[FohmComplementaryReport] = fohm_upload_api.get_sars_cov_complementary_reports(
        fohm_complementary_reports
    )

    # THEN a list of reports is returned
    assert isinstance(content[0], FohmComplementaryReport)

    # THEN only the report for Sars-cov2 reports remains
    assert len(content) == 1
    assert content[0].sample_number == "44CS000000"


def test_get_sars_cov_pangolin_reports(
    fohm_pangolin_reports: list[FohmPangolinReport], fohm_upload_api: FOHMUploadAPI
):
    # GIVEN a list of reports

    # WHEN getting Sars-cov reports from reports
    content: list[FohmPangolinReport] = fohm_upload_api.get_sars_cov_pangolin_reports(
        fohm_pangolin_reports
    )

    # THEN a list of reports is returned
    assert isinstance(content[0], FohmPangolinReport)

    # THEN only the report for Sars-cov2 reports remains
    assert len(content) == 1
    assert content[0].taxon == "44CS000001"


def test_add_sample_internal_id_to_complementary_reports(
    fohm_complementary_reports: list[FohmComplementaryReport], fohm_upload_api: FOHMUploadAPI
):
    """Test adding sample internal ids to the reports."""
    # GIVEN a FOHM upload API

    # GIVEN a list of complementary reports

    # WHEN adding sample internal ids to reports
    fohm_upload_api.add_sample_internal_id_to_complementary_reports(fohm_complementary_reports)

    # THEN a sample internal id has been added
    assert isinstance(fohm_complementary_reports[0].internal_id, str)


def test_add_sample_internal_id_to_pangolin_reports(
    fohm_pangolin_reports: list[FohmPangolinReport], fohm_upload_api: FOHMUploadAPI
):
    """Test adding sample internal ids to the reports."""
    # GIVEN a FOHM upload API

    # GIVEN a list of Pangolin reports

    # WHEN adding sample internal ids to reports
    fohm_upload_api.add_sample_internal_id_to_pangolin_reports(fohm_pangolin_reports)

    # THEN a sample internal id has been added
    assert isinstance(fohm_pangolin_reports[0].internal_id, str)


def test_add_region_lab_to_reports(
    fohm_pangolin_reports: list[FohmPangolinReport], fohm_upload_api: FOHMUploadAPI
):
    """Test adding region laboratories to the reports."""
    # GIVEN a FOHM upload API

    # GIVEN a list of Pangolin reports

    # WHEN adding region lab to reports
    fohm_upload_api.add_region_lab_to_reports(fohm_pangolin_reports)

    # THEN a region lab has been added
    assert isinstance(fohm_pangolin_reports[0].region_lab, str)


def test_create_pangolin_reports_csv(
    fohm_pangolin_reports: list[FohmPangolinReport], fohm_upload_api: FOHMUploadAPI
):
    # GIVEN a list of reports

    # GIVEN report folders exist
    fohm_upload_api.create_daily_delivery_folders()

    # GIVEN an outdata path for reports
    pangolin_report_file = Path(
        fohm_upload_api.daily_rawdata_path,
        f"None_{fohm_upload_api.current_datestr}_pangolin_classification_format4{FileExtensions.TXT}",
    )

    # GIVEN that the Pangolin report path does not exist
    assert not pangolin_report_file.exists()

    # WHEN creating reports
    fohm_upload_api.create_pangolin_report(fohm_pangolin_reports)

    # THEN a file is generated
    assert pangolin_report_file.exists()


def test_create_complementary_report(
    fohm_complementary_reports: list[FohmComplementaryReport], fohm_upload_api: FOHMUploadAPI
):
    # GIVEN a list of reports

    # GIVEN report folders exist
    fohm_upload_api.create_daily_delivery_folders()

    # GIVEN an outdata path for reports
    complementary_report_file = Path(
        fohm_upload_api.daily_report_path,
        f"None_{fohm_upload_api.current_datestr}_komplettering{FileExtensions.CSV}",
    )

    # GIVEN that the complementary report path does not exist
    assert not complementary_report_file.exists()

    # WHEN creating reports
    fohm_upload_api.create_complementary_report(fohm_complementary_reports)

    # THEN a file is generated
    assert complementary_report_file.exists()



import logging
import shutil
from pathlib import Path
from cg.constants import SequencingFileTag, Workflow
from cg.constants.constants import FileFormat, MutantQC
from cg.constants.tb import AnalysisStatus
from cg.exc import CgError
from cg.io.controller import WriteFile
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.fastq import MutantFastqHandler
from cg.services.sequencing_qc_service.sequencing_qc_service import SequencingQCService
from cg.meta.workflow.mutant.quality_controller.models import MutantQualityResult
from cg.meta.workflow.mutant.quality_controller.quality_controller import MutantQualityController
from cg.models.cg_config import CGConfig
from cg.models.workflow.mutant import MutantSampleConfig
from cg.store.models import Application, Case, Sample
from cg.utils import Process

LOG = logging.getLogger(__name__)


class MutantAnalysisAPI(AnalysisAPI):
    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.MUTANT,
    ):
        super().__init__(workflow=workflow, config=config)
        self.root_dir = config.mutant.root
        self.quality_checker = MutantQualityController(
            status_db=config.status_db, lims=config.lims_api
        )

    @property
    def conda_binary(self) -> str:
        return self.config.mutant.conda_binary

    @property
    def process(self) -> Process:
        if not self._process:
            self._process = Process(
                binary=self.config.mutant.binary_path,
                conda_binary=f"{self.conda_binary}" if self.conda_binary else None,
                environment=self.config.mutant.conda_env,
            )
        return self._process

    @property
    def fastq_handler(self):
        return MutantFastqHandler

    def get_case_path(self, case_id: str) -> Path:
        return Path(self.root_dir, case_id)

    def get_case_output_path(self, case_id: str) -> Path:
        return Path(self.get_case_path(case_id=case_id), "results")

    def get_case_results_file_path(self, case: Case) -> Path:
        case_output_path: Path = self.get_case_output_path(case.internal_id)
        return Path(case_output_path, f"sars-cov-2_{case.latest_ticket}_results.csv")

    def get_case_fastq_dir(self, case_id: str) -> Path:
        return Path(self.get_case_path(case_id=case_id), "fastq")

    def get_case_qc_report_path(self, case_id: str) -> Path:
        case_path: Path = self.get_case_path(case_id=case_id)
        return Path(case_path, MutantQC.QUALITY_REPORT_FILE_NAME)

    def get_job_ids_path(self, case_id: str) -> Path:
        return Path(self.get_case_output_path(case_id=case_id), "trailblazer_config.yaml")

    def _is_nanopore(self, application: Application) -> bool:
        return application.tag[3:6] == "ONT"

    def get_sample_fastq_destination_dir(self, case: Case, sample: Sample) -> Path:
        """Return the path to the FASTQ destination directory."""
        application: str = sample.application_version.application
        if self._is_nanopore(application=application):
            return Path(self.get_case_path(case_id=case.internal_id), FileFormat.FASTQ, sample.name)
        return Path(self.get_case_path(case_id=case.internal_id), FileFormat.FASTQ)

    def get_case_config_path(self, case_id: str) -> Path:
        return Path(self.get_case_path(case_id=case_id), "case_config.json")

    def get_deliverables_file_path(self, case_id: str) -> Path:
        return Path(self.get_case_output_path(case_id=case_id), f"{case_id}_deliverables.yaml")

    def link_fastq_files(self, case_id: str, dry_run: bool = False) -> None:
        case_obj = self.status_db.get_case_by_internal_id(internal_id=case_id)
        samples: list[Sample] = [link.sample for link in case_obj.links]
        for sample_obj in samples:
            application = sample_obj.application_version.application
            if self._is_nanopore(application):
                self.link_nanopore_fastq_for_sample(
                    case=case_obj, sample=sample_obj, concatenate=True
                )
                continue
            if not SequencingQCService.sample_pass_sequencing_qc(sample_obj):
                LOG.info(f"Sample {sample_obj.internal_id} read count below threshold, skipping!")
                continue
            self.link_fastq_files_for_sample(case=case_obj, sample=sample_obj, concatenate=True)

    def get_sample_parameters(self, sample_obj: Sample, sequencing_qc: bool) -> MutantSampleConfig:
        return MutantSampleConfig(
            CG_ID_sample=sample_obj.internal_id,
            case_ID=sample_obj.links[0].case.internal_id,
            Customer_ID_sample=sample_obj.name,
            customer_id=sample_obj.customer.internal_id,
            sequencing_qc_pass=sequencing_qc,
            CG_ID_project=sample_obj.links[0].case.name,
            Customer_ID_project=sample_obj.original_ticket,
            application_tag=sample_obj.application_version.application.tag,
            method_libprep=str(self.lims_api.get_prep_method(lims_id=sample_obj.internal_id)),
            method_sequencing=str(
                self.lims_api.get_sequencing_method(lims_id=sample_obj.internal_id)
            ),
            date_arrival=str(sample_obj.received_at),
            date_libprep=str(sample_obj.prepared_at),
            date_sequencing=str(sample_obj.last_sequenced_at),
            selection_criteria=self.lims_api.get_sample_attribute(
                lims_id=sample_obj.internal_id, key="selection_criteria"
            ),
            priority=self.lims_api.get_sample_attribute(
                lims_id=sample_obj.internal_id, key="priority"
            ),
            region_code=self.lims_api.get_sample_attribute(
                lims_id=sample_obj.internal_id, key="region_code"
            ),
            lab_code=self.lims_api.get_sample_attribute(
                lims_id=sample_obj.internal_id, key="lab_code"
            ),
            primer=self.lims_api.get_sample_attribute(lims_id=sample_obj.internal_id, key="primer"),
        )

    def create_case_config(self, case_id: str, dry_run: bool) -> None:
        case_obj = self.status_db.get_case_by_internal_id(internal_id=case_id)
        samples: list[Sample] = [link.sample for link in case_obj.links]
        case_config_list: list[dict] = []
        for sample in samples:
            sample_passed_sequencing_qc: bool = SequencingQCService.sample_pass_sequencing_qc(
                sample=sample
            )
            case_config_list.append(
                self.get_sample_parameters(
                    sample_obj=sample, sequencing_qc=sample_passed_sequencing_qc
                ).model_dump()
            )
        config_path: Path = self.get_case_config_path(case_id=case_id)
        if dry_run:
            LOG.info(
                f"Dry-run, would have created config at path {config_path.as_posix()}, with content:"
            )
            LOG.info(case_config_list)
        config_path.parent.mkdir(parents=True, exist_ok=True)
        WriteFile.write_file_from_content(
            content=case_config_list, file_format=FileFormat.JSON, file_path=config_path
        )
        LOG.info(f"Saved config to {config_path.as_posix()}")

    def get_lims_naming_metadata(self, sample: Sample) -> str | None:
        region_code = self.lims_api.get_sample_attribute(
            lims_id=sample.internal_id, key="region_code"
        ).split(" ")[0]
        lab_code = self.lims_api.get_sample_attribute(
            lims_id=sample.internal_id, key="lab_code"
        ).split(" ")[0]

        return f"{region_code}_{lab_code}_{sample.name}"

    def run_analysis(self, case_id: str, dry_run: bool, config_artic: str = None) -> None:
        if self.get_case_output_path(case_id=case_id).exists():
            LOG.info("Found old output files, directory will be cleaned!")
            shutil.rmtree(self.get_case_output_path(case_id=case_id), ignore_errors=True)

        if config_artic:
            self.process.run_command(
                [
                    "analyse",
                    "sarscov2",
                    "--config_case",
                    self.get_case_config_path(case_id=case_id).as_posix(),
                    "--config_artic",
                    config_artic,
                    "--outdir",
                    self.get_case_output_path(case_id=case_id).as_posix(),
                    self.get_case_fastq_dir(case_id=case_id).as_posix(),
                ],
                dry_run=dry_run,
            )
        else:
            self.process.run_command(
                [
                    "analyse",
                    "sarscov2",
                    "--config_case",
                    self.get_case_config_path(case_id=case_id).as_posix(),
                    "--outdir",
                    self.get_case_output_path(case_id=case_id).as_posix(),
                    self.get_case_fastq_dir(case_id=case_id).as_posix(),
                ],
                dry_run=dry_run,
            )

    def get_cases_to_store(self) -> list[Case]:
        """Return cases for which the analysis is complete on Traiblazer and a QC report has been generated."""
        cases_to_store: list[Case] = [
            case
            for case in self.status_db.get_running_cases_in_workflow(self.workflow)
            if self.trailblazer_api.is_latest_analysis_completed(case.internal_id)
            and self.get_case_qc_report_path(case_id=case.internal_id).exists()
        ]
        return cases_to_store

    def get_cases_to_perform_qc_on(self) -> list[Case]:
        """Return cases with a completed analysis that are not yet stored."""
        cases_to_perform_qc_on: list[Case] = [
            case
            for case in self.status_db.get_running_cases_in_workflow(self.workflow)
            if self.trailblazer_api.is_latest_analysis_completed(case.internal_id)
            and not self.get_case_qc_report_path(case_id=case.internal_id).exists()
        ]
        return cases_to_perform_qc_on

    def get_metadata_for_nanopore_sample(self, sample: Sample) -> list[dict]:
        return [
            self.fastq_handler.parse_nanopore_file_data(file_obj.full_path)
            for file_obj in self.housekeeper_api.files(
                bundle=sample.internal_id, tags={SequencingFileTag.FASTQ}
            )
        ]

    def link_nanopore_fastq_for_sample(
        self, case: Case, sample: Sample, concatenate: bool = False
    ) -> None:
        """
        Link FASTQ files for a nanopore sample to working directory.
        If workflow input requires concatenated fastq, files can also be concatenated
        """
        read_paths = []
        files: list[dict] = self.get_metadata_for_nanopore_sample(sample=sample)
        sorted_files = sorted(files, key=lambda k: k["path"])
        fastq_dir = self.get_sample_fastq_destination_dir(case=case, sample=sample)
        fastq_dir.mkdir(parents=True, exist_ok=True)

        for counter, fastq_data in enumerate(sorted_files):
            fastq_path = Path(fastq_data["path"])
            fastq_name = self.fastq_handler.create_nanopore_fastq_name(
                flowcell=fastq_data["flowcell"],
                sample=sample.internal_id,
                filenr=str(counter),
                meta=self.get_lims_naming_metadata(sample),
            )
            destination_path: Path = fastq_dir / fastq_name
            read_paths.append(destination_path)

            if not destination_path.exists():
                LOG.info(f"Linking: {fastq_path} -> {destination_path}")
                destination_path.symlink_to(fastq_path)
            else:
                LOG.warning(f"Destination path already exists: {destination_path}")

        if not concatenate:
            return

        concatenated_fastq_name = self.fastq_handler.create_nanopore_fastq_name(
            flowcell=fastq_data["flowcell"],
            sample=sample.internal_id,
            filenr=str(1),
            meta=self.get_lims_naming_metadata(sample),
        )
        concatenated_path = (
            f"{fastq_dir}/{self.fastq_handler.get_concatenated_name(concatenated_fastq_name)}"
        )
        LOG.info(f"Concatenation in progress for sample {sample.internal_id}.")
        self.fastq_handler.concatenate(read_paths, concatenated_path)
        self.fastq_handler.remove_files(read_paths)

    def run_qc_on_case(self, case: Case, dry_run: bool) -> None:
        """Run qc check on case, report qc summary on Trailblazer and set analysis status to fail if it fails QC."""
        try:
            qc_result: MutantQualityResult = self.get_qc_result(case=case)
        except Exception as exception:
            error_message: str = f"Could not perform QC on case {case.internal_id}: {exception}"
            LOG.error(error_message)
            if not dry_run:
                self.trailblazer_api.add_comment(
                    case_id=case.internal_id, comment="ERROR: Could not perform QC on case"
                )
                self.trailblazer_api.set_analysis_status(
                    case_id=case.internal_id, status=AnalysisStatus.ERROR
                )
            raise CgError(error_message)

        if not dry_run:
            self.report_qc_on_trailblazer(case=case, qc_result=qc_result)
            if not qc_result.passes_qc:
                self.trailblazer_api.set_analysis_status(
                    case_id=case.internal_id, status=AnalysisStatus.FAILED
                )

    def get_qc_result(self, case: Case) -> MutantQualityResult:
        case_results_file_path: Path = self.get_case_results_file_path(case=case)
        case_qc_report_path: Path = self.get_case_qc_report_path(case_id=case.internal_id)
        qc_result: MutantQualityResult = self.quality_checker.get_quality_control_result(
            case=case,
            case_results_file_path=case_results_file_path,
            case_qc_report_path=case_qc_report_path,
        )
        return qc_result

    def report_qc_on_trailblazer(self, case: Case, qc_result: MutantQualityResult) -> None:
        report_file_path: Path = self.get_case_qc_report_path(case_id=case.internal_id)

        comment = qc_result.summary + (
            f" QC report: {report_file_path}" if not qc_result.passes_qc else ""
        )
        self.trailblazer_api.add_comment(case_id=case.internal_id, comment=comment)

    def run_qc(self, case_id: str, dry_run: bool) -> None:
        LOG.info(f"Running QC on case {case_id}.")

        case: Case = self.status_db.get_case_by_internal_id(case_id)

        self.run_qc_on_case(case=case, dry_run=dry_run)



import os
from pathlib import Path
from cg.constants.delivery import INBOX_NAME
from cg.services.deliver_files.delivery_file_fetcher_service.models import (
    DeliveryFiles,
    SampleFile,
    CaseFile,
)
from cg.services.deliver_files.delivery_file_formatter_service.utils.case_file_formatter import (
    CaseFileFormatter,
)
from cg.services.deliver_files.delivery_file_formatter_service.delivery_file_formatting_service import (
    DeliveryFileFormattingService,
)
from cg.services.deliver_files.delivery_file_formatter_service.models import (
    FormattedFiles,
    FormattedFile,
)
from cg.services.deliver_files.delivery_file_formatter_service.utils.sample_file_concatenation_formatter import (
    SampleFileConcatenationFormatter,
)
from cg.services.deliver_files.delivery_file_formatter_service.utils.sample_file_formatter import (
    SampleFileFormatter,
)


class DeliveryFileFormatter(DeliveryFileFormattingService):
    """
    Format the files to be delivered in the generic format.
    Expected structure:
    <customer>/inbox/<ticket_id>/<case_name>/<case_files>
    <customer>/inbox/<ticket_id>/<sample_name>/<sample_files>
    """

    def __init__(
        self,
        case_file_formatter: CaseFileFormatter,
        sample_file_formatter: SampleFileFormatter | SampleFileConcatenationFormatter,
    ):
        self.case_file_formatter = case_file_formatter
        self.sample_file_formatter = sample_file_formatter

    def format_files(self, delivery_files: DeliveryFiles) -> FormattedFiles:
        """Format the files to be delivered and return the formatted files in the generic format."""
        ticket_dir_path: Path = self.get_folder_under_inbox(
            delivery_files.sample_files[0].file_path
        )
        self._create_ticket_dir(ticket_dir_path)
        formatted_files: list[FormattedFile] = self._format_sample_and_case_files(
            sample_files=delivery_files.sample_files,
            case_files=delivery_files.case_files,
            ticket_dir_path=ticket_dir_path,
        )
        return FormattedFiles(files=formatted_files)

    def _format_sample_and_case_files(
        self, sample_files: list[SampleFile], case_files: list[CaseFile], ticket_dir_path: Path
    ) -> list[FormattedFile]:
        """Helper method to format both sample and case files."""
        formatted_files: list[FormattedFile] = self.sample_file_formatter.format_files(
            moved_files=sample_files,
            ticket_dir_path=ticket_dir_path,
        )
        if case_files:
            formatted_case_files: list[FormattedFile] = self.case_file_formatter.format_files(
                moved_files=case_files,
                ticket_dir_path=ticket_dir_path,
            )
            formatted_files.extend(formatted_case_files)
        return formatted_files

    @staticmethod
    def get_folder_under_inbox(file_path: Path) -> Path:
        """Get the ticker folder that is located under the customer inbox folder."""
        try:
            inbox_index: int = file_path.parts.index(INBOX_NAME)
            return Path(*file_path.parts[: inbox_index + 2])
        except ValueError:
            raise ValueError(f"Could not find the inbox directory in the path: {file_path}")

    @staticmethod
    def _create_ticket_dir(ticket_dir_path: Path) -> None:
        os.makedirs(ticket_dir_path, exist_ok=True)



import logging
from functools import wraps

from cg.exc import HousekeeperBundleVersionMissingError

LOG = logging.getLogger(__name__)


def handle_missing_bundle_errors(func):
    """
    Log an error when a Housekeeper bundle version is missing.

    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except HousekeeperBundleVersionMissingError as error:
            LOG.error(error)
            return None

    return wrapper



"""Module for the data transfer models for Illumina."""

from datetime import datetime

from pydantic import BaseModel, field_validator

from cg.constants import SequencingRunDataAvailability
from cg.constants.devices import DeviceType
from cg.constants.sequencing import Sequencers
from cg.utils.calculations import fraction_to_percent


class IlluminaFlowCellDTO(BaseModel):
    """Data transfer object for Illumina flow cell."""

    internal_id: str
    type: DeviceType
    model: str | None

    class Config:
        arbitrary_types_allowed = True

    @field_validator("model")
    def validate_model(cls, model: str) -> str:
        """Validate the model."""
        if model not in ["10B", "25B", "1.5B", "S1", "S2", "S4", "SP", None]:
            raise ValueError(f"Invalid Flow cell model detected: {model}")
        return model


class IlluminaSequencingRunDTO(BaseModel):
    """Data transfer object for IlluminaSequencingRun."""

    type: DeviceType
    sequencer_type: Sequencers | None
    sequencer_name: str | None
    data_availability: SequencingRunDataAvailability | None
    archived_at: datetime | None
    has_backup: bool | None
    total_reads: int | None
    total_undetermined_reads: int | None
    percent_undetermined_reads: float | None
    percent_q30: float | None
    mean_quality_score: float | None
    total_yield: int | None
    yield_q30: int | None
    cycles: int | None
    demultiplexing_software: str | None
    demultiplexing_software_version: str | None
    sequencing_started_at: datetime | None
    sequencing_completed_at: datetime | None
    demultiplexing_started_at: datetime | None
    demultiplexing_completed_at: datetime | None

    @field_validator("percent_undetermined_reads", "percent_q30", mode="before")
    def validate_percent_fields(cls, value: float | None) -> float | None:
        if value is not None:
            return fraction_to_percent(value)
        return value

    class Config:
        arbitrary_types_allowed = True
        validate_assignment = True


class IlluminaSampleSequencingMetricsDTO(BaseModel):
    """Data transfer object for IlluminaSampleSequencingMetrics."""

    sample_id: str
    type: DeviceType
    flow_cell_lane: int
    total_reads_in_lane: int
    base_passing_q30_percent: float
    base_mean_quality_score: float
    yield_: int
    yield_q30: float
    created_at: datetime

    @field_validator("base_passing_q30_percent", mode="before")
    def validate_percent_fields(cls, value: float | None) -> float | None:
        if value is not None:
            return fraction_to_percent(value)
        return value

    class Config:
        arbitrary_types_allowed = True
        validate_assignment = True



from pathlib import Path

import pytest

from cg.services.deliver_files.delivery_file_fetcher_service.models import DeliveryFiles, SampleFile
from cg.services.deliver_files.delivery_file_formatter_service.models import FormattedFile


@pytest.fixture
def expected_formatted_analysis_case_files(
    expected_moved_analysis_delivery_files: DeliveryFiles,
) -> list[FormattedFile]:
    formatted_files: list[FormattedFile] = []
    for case_file in expected_moved_analysis_delivery_files.case_files:
        replaced_case_file_name: str = case_file.file_path.name.replace(
            case_file.case_id, case_file.case_name
        )
        formatted_file_path = Path(
            case_file.file_path.parent, case_file.case_name, replaced_case_file_name
        )
        formatted_files.append(
            FormattedFile(original_path=case_file.file_path, formatted_path=formatted_file_path)
        )
    return formatted_files


@pytest.fixture
def expected_formatted_analysis_sample_files(
    expected_moved_analysis_delivery_files: DeliveryFiles,
) -> list[FormattedFile]:
    formatted_files: list[FormattedFile] = []
    for sample_file in expected_moved_analysis_delivery_files.sample_files:
        replaced_sample_file_name: str = sample_file.file_path.name.replace(
            sample_file.sample_id, sample_file.sample_name
        )
        formatted_file_path = Path(
            sample_file.file_path.parent, sample_file.sample_name, replaced_sample_file_name
        )
        formatted_files.append(
            FormattedFile(original_path=sample_file.file_path, formatted_path=formatted_file_path)
        )
    return formatted_files


@pytest.fixture
def expected_formatted_fastq_sample_files(
    expected_moved_fastq_delivery_files: DeliveryFiles,
) -> list[FormattedFile]:
    formatted_files: list[FormattedFile] = []
    for sample_file in expected_moved_fastq_delivery_files.sample_files:
        replaced_sample_file_name: str = sample_file.file_path.name.replace(
            sample_file.sample_id, sample_file.sample_name
        )
        formatted_file_path = Path(
            sample_file.file_path.parent, sample_file.sample_name, replaced_sample_file_name
        )
        formatted_files.append(
            FormattedFile(original_path=sample_file.file_path, formatted_path=formatted_file_path)
        )
    return formatted_files


@pytest.fixture
def expected_concatenated_fastq_formatted_files(
    fastq_concatenation_sample_files,
) -> list[FormattedFile]:
    formatted_files: list[FormattedFile] = []
    for sample_file in fastq_concatenation_sample_files:
        replaced_sample_file_name: str = sample_file.file_path.name.replace(
            sample_file.sample_id, sample_file.sample_name
        )
        replaced_sample_file_name = replaced_sample_file_name.replace("1_R1_1", "1")
        replaced_sample_file_name = replaced_sample_file_name.replace("2_R1_1", "1")
        replaced_sample_file_name = replaced_sample_file_name.replace("1_R2_1", "2")
        replaced_sample_file_name = replaced_sample_file_name.replace("2_R2_1", "2")
        formatted_file_path = Path(
            sample_file.file_path.parent, sample_file.sample_name, replaced_sample_file_name
        )
        formatted_files.append(
            FormattedFile(original_path=sample_file.file_path, formatted_path=formatted_file_path)
        )
    return formatted_files


@pytest.fixture
def empty_case_files() -> None:
    return None



"""Nf-tower related constants."""

from enum import StrEnum
from typing import Any


class NfTowerStatus(StrEnum):
    """NF-Tower job submission status."""

    SUBMITTED: str = "SUBMITTED"
    RUNNING: str = "RUNNING"
    SUCCEEDED: str = "SUCCEEDED"
    FAILED: str = "FAILED"
    CANCELLED: str = "CANCELLED"
    UNKNOWN: str = "UNKNOWN"


RAREDISEASE_METRIC_CONDITIONS: dict[str, dict[str, Any]] = {
    "percent_duplicates": {"norm": "lt", "threshold": 20},
    "PCT_PF_UQ_READS_ALIGNED": {"norm": "gt", "threshold": 0.95},
    "MEDIAN_TARGET_COVERAGE": {"norm": "gt", "threshold": 25},
    "PCT_TARGET_BASES_10X": {"norm": "gt", "threshold": 0.95},
    "PCT_EXC_ADAPTER": {"norm": "lt", "threshold": 0.0005},
    "predicted_sex_sex_check": {"norm": "eq", "threshold": None},
}

RAREDISEASE_PARENT_PEDDY_METRIC_CONDITION: dict[str, dict[str, Any]] = {
    "parent_error_ped_check": {"norm": "eq", "threshold": "False"},
}

RNAFUSION_METRIC_CONDITIONS: dict[str, dict[str, Any]] = {
    "uniquely_mapped_percent": {"norm": "gt", "threshold": 60},
    "PCT_MRNA_BASES": {"norm": "gt", "threshold": 80},
    "PCT_RIBOSOMAL_BASES": {"norm": "lt", "threshold": 5},
    "PERCENT_DUPLICATION": {"norm": "lt", "threshold": 0.7},
}

TOMTE_METRIC_CONDITIONS: dict[str, dict[str, Any]] = {
    "uniquely_mapped_percent": {"norm": "gt", "threshold": 60},
    "PCT_MRNA_BASES": {"norm": "gt", "threshold": 80},
    "PCT_RIBOSOMAL_BASES": {"norm": "lt", "threshold": 5},
    "pct_duplication": {"norm": "lt", "threshold": 70},
}


MULTIQC_NEXFLOW_CONFIG = """process {
    withName:'MULTIQC' {
        memory = { 1.GB * task.attempt }
        time   = { 4.h  * task.attempt }
        cpus = 2
        ext.args = ' --data-format json '
    }
}
"""

RAREDISEASE_COVERAGE_FILE_TAGS: list[str] = ["coverage", "d4"]
RAREDISEASE_COVERAGE_INTERVAL_TYPE: str = "genes"
RAREDISEASE_COVERAGE_THRESHOLD: int = 10



"""This API should handle everything around demultiplexing."""

import logging
import shutil
from pathlib import Path

from typing_extensions import Literal

from cg.apps.demultiplex.sbatch import DEMULTIPLEX_COMMAND, DEMULTIPLEX_ERROR
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.slurm.slurm_api import SlurmAPI
from cg.apps.tb import TrailblazerAPI
from cg.constants.constants import FileFormat, Workflow
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.constants.priority import SlurmQos
from cg.constants.tb import AnalysisTypes
from cg.exc import HousekeeperFileMissingError
from cg.io.controller import WriteFile
from cg.models.demultiplex.sbatch import SbatchCommand, SbatchError
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.models.slurm.sbatch import SbatchDragen

LOG = logging.getLogger(__name__)


class DemultiplexingAPI:
    """Demultiplexing API should deal with anything related to demultiplexing.

    This includes starting demultiplexing, creating sample sheets, creating base masks,
    """

    def __init__(self, config: dict, housekeeper_api: HousekeeperAPI, out_dir: Path | None = None):
        self.slurm_api = SlurmAPI()
        self.hk_api = housekeeper_api
        self.slurm_account: str = config["demultiplex"]["slurm"]["account"]
        self.mail: str = config["demultiplex"]["slurm"]["mail_user"]
        self.sequencing_runs_dir: Path = Path(
            config["run_instruments"]["illumina"]["sequencing_runs_dir"]
        )
        self.demultiplexed_runs_dir: Path = out_dir or Path(
            config["run_instruments"]["illumina"]["demultiplexed_runs_dir"]
        )
        self.environment: str = config.get("environment", "stage")
        LOG.info(f"Set environment to {self.environment}")
        self.dry_run: bool = False

    @property
    def slurm_quality_of_service(self) -> Literal[SlurmQos.HIGH, SlurmQos.LOW]:
        """Return SLURM quality of service."""
        return SlurmQos.LOW if self.environment == "stage" else SlurmQos.HIGH

    def set_dry_run(self, dry_run: bool) -> None:
        """Set dry run."""
        LOG.debug(f"DemultiplexingAPI: Set dry run to {dry_run}")
        self.dry_run = dry_run
        self.slurm_api.set_dry_run(dry_run=dry_run)

    @staticmethod
    def get_sbatch_error(
        sequencing_run: IlluminaRunDirectoryData,
        email: str,
        demux_dir: Path,
    ) -> str:
        """Create the sbatch error string."""
        LOG.debug("Creating the sbatch error string")
        error_parameters: SbatchError = SbatchError(
            flow_cell_id=sequencing_run.id,
            email=email,
            logfile=DemultiplexingAPI.get_stderr_logfile(sequencing_run=sequencing_run).as_posix(),
            demux_dir=demux_dir.as_posix(),
            demux_started=sequencing_run.demultiplexing_started_path.as_posix(),
        )
        return DEMULTIPLEX_ERROR.format(**error_parameters.model_dump())

    @staticmethod
    def get_sbatch_command(
        run_dir: Path,
        demux_dir: Path,
        sample_sheet: Path,
        demux_completed: Path,
        environment: Literal["production", "stage"] = "stage",
    ) -> str:
        """
        Return sbatch command.
        The unaligned_dir is only used for Bcl2Fastq.
        """
        LOG.debug("Creating the sbatch command string")
        unaligned_dir: Path = Path(demux_dir, DemultiplexingDirsAndFiles.UNALIGNED_DIR_NAME)
        command_parameters: SbatchCommand = SbatchCommand(
            run_dir=run_dir.as_posix(),
            demux_dir=demux_dir.as_posix(),
            unaligned_dir=unaligned_dir.as_posix(),
            sample_sheet=sample_sheet.as_posix(),
            demux_completed_file=demux_completed.as_posix(),
            environment=environment,
        )
        return DEMULTIPLEX_COMMAND.format(**command_parameters.model_dump())

    @staticmethod
    def demultiplex_sbatch_path(sequencing_run: IlluminaRunDirectoryData) -> Path:
        """Get the path to where sbatch script file should be kept."""
        return Path(sequencing_run.path, "demux-novaseq.sh")

    @staticmethod
    def get_run_name(sequencing_run: IlluminaRunDirectoryData) -> str:
        """Create the run name for the sbatch job."""
        return f"{sequencing_run.id}_demultiplex"

    @staticmethod
    def get_stderr_logfile(sequencing_run: IlluminaRunDirectoryData) -> Path:
        """Create the path to the stderr logfile."""
        return Path(sequencing_run.path, f"{DemultiplexingAPI.get_run_name(sequencing_run)}.stderr")

    @staticmethod
    def get_stdout_logfile(sequencing_run: IlluminaRunDirectoryData) -> Path:
        """Create the path to the stdout logfile."""
        return Path(sequencing_run.path, f"{DemultiplexingAPI.get_run_name(sequencing_run)}.stdout")

    def demultiplexed_run_dir_path(self, sequencing_run: IlluminaRunDirectoryData) -> Path:
        """Create the path to where the demultiplexed result should be produced."""
        return Path(self.demultiplexed_runs_dir, sequencing_run.path.name)

    def is_sample_sheet_in_housekeeper(self, flow_cell_id: str) -> bool:
        """Returns True if the sample sheet for the flow cell exists in Housekeeper."""
        try:
            self.hk_api.get_sample_sheet_path(flow_cell_id)
            return True
        except HousekeeperFileMissingError:
            return False

    def get_sequencing_run_unaligned_dir(self, sequencing_run: IlluminaRunDirectoryData) -> Path:
        """Returns the path to where the demultiplexed result are located."""
        return Path(
            self.demultiplexed_run_dir_path(sequencing_run),
            DemultiplexingDirsAndFiles.UNALIGNED_DIR_NAME,
        )

    def demultiplexing_completed_path(self, sequencing_run: IlluminaRunDirectoryData) -> Path:
        """Return the path to demultiplexing complete file."""
        LOG.info(
            Path(
                self.demultiplexed_run_dir_path(sequencing_run),
                DemultiplexingDirsAndFiles.DEMUX_COMPLETE,
            )
        )
        return Path(
            self.demultiplexed_run_dir_path(sequencing_run),
            DemultiplexingDirsAndFiles.DEMUX_COMPLETE,
        )

    def is_demultiplexing_possible(self, sequencing_run: IlluminaRunDirectoryData) -> bool:
        """Check if it is possible to start demultiplexing.

        This means that
            - sequencing run should be ready for demultiplexing (all files in place)
            - sample sheet needs to exist
            - demultiplexing should not be running
        """
        LOG.info(f"Check if demultiplexing is possible for {sequencing_run.id}")
        demultiplexing_possible = True
        if not sequencing_run.is_sequencing_run_ready():
            demultiplexing_possible = False

        if not sequencing_run.sample_sheet_exists():
            LOG.warning(
                f"Could not find sample sheet in sequencing run directory for {sequencing_run.id}"
            )
            demultiplexing_possible = False

        if not self.is_sample_sheet_in_housekeeper(flow_cell_id=sequencing_run.id):
            LOG.warning(f"Could not find sample sheet in Housekeeper for {sequencing_run.id}")
            demultiplexing_possible = False

        if (
            sequencing_run.has_demultiplexing_started_locally()
            or sequencing_run.has_demultiplexing_started_on_sequencer()
        ):
            LOG.warning("Demultiplexing has already been started")
            demultiplexing_possible = False

        return demultiplexing_possible

    def create_demultiplexing_started_file(self, demultiplexing_started_path: Path) -> None:
        LOG.info("Creating demultiplexing started file")
        if self.dry_run:
            return
        demultiplexing_started_path.touch(exist_ok=False)

    @staticmethod
    def get_trailblazer_config(slurm_job_id: int) -> dict[str, list[str]]:
        return {"jobs": [str(slurm_job_id)]}

    @staticmethod
    def write_trailblazer_config(content: dict, file_path: Path) -> None:
        """Write the content to a yaml file"""
        LOG.info(f"Writing yaml content {content} to {file_path}")
        WriteFile.write_file_from_content(
            content=content, file_format=FileFormat.YAML, file_path=file_path
        )

    def add_to_trailblazer(
        self, tb_api: TrailblazerAPI, slurm_job_id: int, sequencing_run: IlluminaRunDirectoryData
    ):
        """Add demultiplexing entry to trailblazer."""
        if self.dry_run:
            return
        self.write_trailblazer_config(
            content=self.get_trailblazer_config(slurm_job_id=slurm_job_id),
            file_path=sequencing_run.trailblazer_config_path,
        )
        tb_api.add_pending_analysis(
            case_id=sequencing_run.id,
            analysis_type=AnalysisTypes.OTHER,
            config_path=sequencing_run.trailblazer_config_path.as_posix(),
            out_dir=sequencing_run.trailblazer_config_path.parent.as_posix(),
            slurm_quality_of_service=self.slurm_quality_of_service,
            email=self.mail,
            workflow=Workflow.DEMULTIPLEX,
        )

    def start_demultiplexing(self, sequencing_run: IlluminaRunDirectoryData):
        """Start demultiplexing for a sequencing run."""
        self.create_demultiplexing_started_file(sequencing_run.demultiplexing_started_path)
        log_path: Path = self.get_stderr_logfile(sequencing_run=sequencing_run)
        error_function: str = self.get_sbatch_error(
            sequencing_run=sequencing_run,
            email=self.mail,
            demux_dir=self.demultiplexed_run_dir_path(sequencing_run),
        )
        commands: str = self.get_sbatch_command(
            run_dir=sequencing_run.path,
            demux_dir=self.demultiplexed_run_dir_path(sequencing_run=sequencing_run),
            sample_sheet=sequencing_run.sample_sheet_path,
            demux_completed=self.demultiplexing_completed_path(sequencing_run=sequencing_run),
            environment=self.environment,
        )
        sbatch_parameters: SbatchDragen = SbatchDragen(
            account=self.slurm_account,
            commands=commands,
            email=self.mail,
            error=error_function,
            hours=36,
            job_name=self.get_run_name(sequencing_run),
            log_dir=log_path.parent.as_posix(),
            quality_of_service=self.slurm_quality_of_service,
        )
        sbatch_content: str = self.slurm_api.generate_sbatch_content(
            sbatch_parameters=sbatch_parameters
        )
        sbatch_path: Path = self.demultiplex_sbatch_path(sequencing_run=sequencing_run)
        sbatch_number: int = self.slurm_api.submit_sbatch(
            sbatch_content=sbatch_content, sbatch_path=sbatch_path
        )
        LOG.info(f"Demultiplexing running as job {sbatch_number}")
        return sbatch_number

    def prepare_output_directory(self, sequencing_run: IlluminaRunDirectoryData) -> None:
        """Makes sure the output directory is ready for demultiplexing."""
        self.remove_demultiplexing_output_directory(sequencing_run)
        self.create_demultiplexing_output_dir(sequencing_run)

    def remove_demultiplexing_output_directory(
        self, sequencing_run: IlluminaRunDirectoryData
    ) -> None:
        if (
            not self.dry_run
            and self.demultiplexed_run_dir_path(sequencing_run=sequencing_run).exists()
        ):
            shutil.rmtree(
                self.demultiplexed_run_dir_path(sequencing_run=sequencing_run), ignore_errors=False
            )

    def create_demultiplexing_output_dir(self, sequencing_run: IlluminaRunDirectoryData) -> None:
        """Creates the demultiplexing output directory for the sequencing run."""
        output_directory: Path = self.demultiplexed_run_dir_path(sequencing_run)
        LOG.debug(f"Creating demultiplexing output directory: {output_directory}")
        output_directory.mkdir(exist_ok=False, parents=True)



from datetime import datetime, timedelta

from sqlalchemy.orm import Query

from cg.constants.constants import Workflow
from cg.store.filters.status_analysis_filters import (
    filter_analyses_by_case_entry_id,
    filter_analyses_by_started_at,
    filter_analyses_not_cleaned,
    filter_analyses_started_before,
    filter_analyses_with_delivery_report,
    filter_analyses_with_workflow,
    filter_analyses_without_delivery_report,
    filter_completed_analyses,
    filter_not_uploaded_analyses,
    filter_report_analyses_by_workflow,
    filter_uploaded_analyses,
    filter_valid_analyses_in_production,
    order_analyses_by_completed_at_asc,
    order_analyses_by_uploaded_at_asc,
    filter_analysis_by_entry_id,
)
from cg.store.models import Analysis, Case
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


def test_filter_valid_analyses_in_production(
    base_store: Store,
    helpers: StoreHelpers,
    case: Case,
    timestamp_now: datetime,
    old_timestamp: datetime,
):
    """Test that an expected analysis is returned when it has a production valid completed_at date."""

    # GIVEN a set of mock analyses
    analysis: Analysis = helpers.add_analysis(store=base_store, completed_at=timestamp_now)
    outdated_analysis: Analysis = helpers.add_analysis(
        store=base_store, case=case, completed_at=old_timestamp
    )

    # WHEN retrieving valid in production analyses
    analyses: Query = filter_valid_analyses_in_production(
        analyses=base_store._get_query(table=Analysis)
    )

    # ASSERT that analyses is a query
    assert isinstance(analyses, Query)

    # THEN only the up-to-date analysis should be returned
    assert analysis in analyses
    assert outdated_analysis not in analyses


def test_filter_analyses_with_workflow(base_store: Store, helpers: StoreHelpers, case: Case):
    """Test analyses filtering by workflow."""

    # GIVEN a set of mock analyses
    balsamic_analysis: Analysis = helpers.add_analysis(store=base_store, workflow=Workflow.BALSAMIC)
    mip_analysis: Analysis = helpers.add_analysis(
        store=base_store, case=case, workflow=Workflow.MIP_DNA
    )

    # WHEN extracting the analyses
    analyses: Query = filter_analyses_with_workflow(
        analyses=base_store._get_query(table=Analysis), workflow=Workflow.BALSAMIC
    )

    # ASSERT that analyses is a query
    assert isinstance(analyses, Query)

    # THEN only the BALSAMIC analysis should be retrieved
    assert balsamic_analysis in analyses
    assert mip_analysis not in analyses


def test_filter_completed_analyses(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test filtering of completed analyses."""

    # GIVEN a mock analysis
    analysis: Analysis = helpers.add_analysis(store=base_store, completed_at=timestamp_now)

    # WHEN retrieving the completed analyses
    analyses: Query = filter_completed_analyses(analyses=base_store._get_query(table=Analysis))

    # ASSERT that analyses is a query
    assert isinstance(analyses, Query)

    # THEN the completed analysis should be obtained
    assert analysis in analyses


def test_filter_filter_uploaded_analyses(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test filtering of analysis with an uploaded_at field."""

    # GIVEN a mock uploaded analysis
    analysis: Analysis = helpers.add_analysis(store=base_store, uploaded_at=timestamp_now)

    # WHEN calling the upload filtering function
    analyses: Query = filter_uploaded_analyses(analyses=base_store._get_query(table=Analysis))

    # ASSERT that analyses is a query
    assert isinstance(analyses, Query)

    # THEN the uploaded analysis should be retrieved
    assert analysis in analyses


def test_filter_not_uploaded_analyses(base_store: Store, helpers: StoreHelpers):
    """Test filtering of analysis that has not been uploaded."""

    # GIVEN a mock not uploaded analysis
    not_uploaded_analysis: Analysis = helpers.add_analysis(store=base_store, uploaded_at=None)

    # WHEN calling the upload filtering function
    analyses: Query = filter_not_uploaded_analyses(analyses=base_store._get_query(table=Analysis))

    # ASSERT that analyses is a query
    assert isinstance(analyses, Query)

    # THEN the uploaded analysis should be retrieved
    assert not_uploaded_analysis in analyses


def test_filter_analyses_with_delivery_report(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test filtering of analysis with a delivery report generated."""

    # GIVEN an analysis with a delivery report
    analysis: Analysis = helpers.add_analysis(store=base_store, delivery_reported_at=timestamp_now)

    # WHEN calling the delivery report analysis filtering function
    analyses: Query = filter_analyses_with_delivery_report(
        analyses=base_store._get_query(table=Analysis)
    )

    # ASSERT that analyses is a query
    assert isinstance(analyses, Query)

    # THEN the analysis containing the delivery report should be extracted
    assert analysis in analyses


def test_filter_analyses_without_delivery_report(base_store: Store, helpers: StoreHelpers):
    """Test filtering of analysis without a delivery report generated."""

    # GIVEN an analysis with a delivery report
    analysis_without_delivery_report: Analysis = helpers.add_analysis(
        store=base_store, delivery_reported_at=None
    )

    # WHEN calling the delivery report analysis filtering function
    analyses: Query = filter_analyses_without_delivery_report(
        analyses=base_store._get_query(table=Analysis)
    )

    # ASSERT that analyses is a query
    assert isinstance(analyses, Query)

    # THEN the analysis without a delivery report should be extracted
    assert analysis_without_delivery_report in analyses


def test_filter_report_analyses_by_workflow(base_store: Store, helpers: StoreHelpers, case: Case):
    """Test filtering delivery report related analysis by workflow."""

    # GIVEN a set of mock analysis
    balsamic_analysis: Analysis = helpers.add_analysis(store=base_store, workflow=Workflow.BALSAMIC)
    fluffy_analysis: Analysis = helpers.add_analysis(
        store=base_store, case=case, workflow=Workflow.FLUFFY
    )

    # WHEN filtering delivery report related analyses
    analyses: Query = filter_report_analyses_by_workflow(
        analyses=base_store._get_query(table=Analysis), workflow=Workflow.BALSAMIC
    )

    # ASSERT that analyses is a query
    assert isinstance(analyses, Query)

    # THEN only the delivery report supported analysis should be retrieved
    assert balsamic_analysis in analyses
    assert fluffy_analysis not in analyses


def test_order_analyses_by_completed_at_asc(
    store: Store,
    helpers: StoreHelpers,
    case: Case,
    timestamp_now: datetime,
    timestamp_yesterday: datetime,
):
    """Test sorting of analyses by the completed_at field."""

    # GIVEN a set of mock analyses
    helpers.add_analysis(store=store, completed_at=timestamp_now)
    helpers.add_analysis(store=store, case=case, completed_at=timestamp_yesterday)

    # WHEN ordering the analyses by the completed_at field
    analyses: Query = order_analyses_by_completed_at_asc(analyses=store._get_query(table=Analysis))

    # ASSERT that analyses is a query
    assert isinstance(analyses, Query)

    # THEN the oldest analysis should be the first one in the list
    for index in range(0, analyses.count() - 1):
        assert analyses.all()[index].completed_at <= analyses.all()[index + 1].completed_at


def test_order_analyses_by_uploaded_at_asc(
    store_with_older_and_newer_analyses: Store,
):
    """Test sorting of analyses by the uploaded_at field."""
    # GIVEN a store with mock analyses

    # WHEN ordering the analyses by the uploaded_at field
    analyses: Query = order_analyses_by_uploaded_at_asc(
        analyses=store_with_older_and_newer_analyses._get_query(table=Analysis)
    )

    # ASSERT that analyses is a query
    assert isinstance(analyses, Query)

    # THEN the oldest analysis should be the first one in the list
    for index in range(0, analyses.count() - 1):
        assert analyses.all()[index].uploaded_at <= analyses.all()[index + 1].uploaded_at


def test_filter_analysis_by_case(base_store: Store, helpers: StoreHelpers, case: Case):
    """Test filtering of analyses by case."""

    # GIVEN a set of mock analyses
    analysis: Analysis = helpers.add_analysis(store=base_store)
    analysis_other_case: Analysis = helpers.add_analysis(store=base_store, case=case)

    # WHEN filtering the analyses by case
    analyses: Query = filter_analyses_by_case_entry_id(
        analyses=base_store._get_query(table=Analysis), case_entry_id=case.id
    )

    # ASSERT that analyses is a query
    assert isinstance(analyses, Query)

    # THEN only the analysis belonging to the case should be retrieved
    assert analysis not in analyses
    assert analysis_other_case in analyses
    assert analysis_other_case.case == case


def test_filter_analysis_started_before(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test filtering of analyses started before a given date."""

    # GIVEN a set of mock analyses
    analysis_old: Analysis = helpers.add_analysis(
        store=base_store, started_at=timestamp_now - timedelta(days=1)
    )
    analysis: Analysis = helpers.add_analysis(
        store=base_store, case=analysis_old.case, started_at=timestamp_now
    )

    # WHEN filtering the analyses by started_at
    analyses: Query = filter_analyses_started_before(
        analyses=base_store._get_query(table=Analysis), started_at_date=timestamp_now
    )

    # ASSERT that analyses is a query
    assert isinstance(analyses, Query)

    # THEN all analyses started before the given date should be retrieved
    for analysis in analyses:
        assert analysis.started_at <= timestamp_now


def test_filter_analysis_not_cleaned(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test filtering of analyses that have not been cleaned."""

    # GIVEN a set of mock analyses
    analysis_cleaned: Analysis = helpers.add_analysis(store=base_store, cleaned_at=timestamp_now)
    analysis: Analysis = helpers.add_analysis(
        store=base_store, case=analysis_cleaned.case, cleaned_at=None
    )

    # WHEN filtering the analyses by cleaned_at
    analyses: Query = filter_analyses_not_cleaned(analyses=base_store._get_query(table=Analysis))

    # ASSERT that analyses is a query
    assert isinstance(analyses, Query)

    # THEN only the analysis that have not been cleaned should be retrieved
    assert analysis in analyses
    assert analysis_cleaned not in analyses


def test_filter_analyses_by_started_at(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime, timestamp_yesterday: datetime
):
    """Test filtering of analyses by started at."""

    # GIVEN a set of mock analyses
    analysis_started_now: Analysis = helpers.add_analysis(
        store=base_store, started_at=timestamp_now
    )
    analysis_started_old: Analysis = helpers.add_analysis(
        store=base_store, case=analysis_started_now.case, started_at=timestamp_yesterday
    )

    # WHEN filtering the analyses by started_at
    analyses: Query = filter_analyses_by_started_at(
        analyses=base_store._get_query(table=Analysis), started_at_date=timestamp_yesterday
    )

    # ASSERT that analyses is a query
    assert isinstance(analyses, Query)

    # THEN only the analysis that have been started after the given date should be retrieved
    assert analysis_started_now not in analyses
    assert analysis_started_old in analyses


def test_filter_by_analysis_entry_id(base_store: Store, helpers: StoreHelpers):
    """Test filtering of analyses by entry id."""

    # GIVEN a set of mock analyses
    analysis: Analysis = helpers.add_analysis(store=base_store)

    # WHEN filtering the analyses by entry id
    analyses: Query = filter_analysis_by_entry_id(
        analyses=base_store._get_query(table=Analysis), entry_id=analysis.id
    )

    # ASSERT that analyses is a query
    assert isinstance(analyses, Query)

    # THEN only the analysis with the given entry id should be retrieved
    assert analysis == analyses.one()



from datetime import datetime

from cg.apps.tb.api import TrailblazerAPI
from cg.apps.tb.models import TrailblazerAnalysis
from cg.constants import Workflow
from cg.store.models import Case, Analysis
from cg.store.store import Store


class AnalysisService:
    def __init__(self, analysis_client: TrailblazerAPI, status_db: Store):
        self.status_db = status_db
        self.analysis_client = analysis_client

    def add_upload_job(self, case_id: str, slurm_id: int):
        analysis: TrailblazerAnalysis = self.analysis_client.get_latest_completed_analysis(case_id)
        self.analysis_client.add_upload_job_to_analysis(slurm_id=slurm_id, analysis_id=analysis.id)

    def get_analyses_to_upload_for_workflow(self, workflow: Workflow) -> list[Analysis]:
        """
        Get all analyses that should be uploaded for a specific workflow.
        1. Get all analyses that should be uploaded for a specific workflow
        2. Check if the analysis has been uploaded and update the uploaded_at field
        """
        analysis_to_upload: list[Analysis] = []
        analyses: list[Analysis] = self.status_db.get_analyses_to_upload(workflow=workflow)
        for analysis in analyses:
            if self._has_uploaded_started(analysis) and self._is_analysis_completed(analysis):
                self.status_db.update_analysis_uploaded_at(
                    analysis_id=analysis.id, uploaded_at=datetime.now()
                )
            if not self._is_analysis_uploaded(analysis) and self._is_analysis_completed(analysis):
                analysis_to_upload.append(analysis)
        return analysis_to_upload

    @staticmethod
    def _is_analysis_uploaded(analysis: Analysis) -> bool:
        return bool(analysis.uploaded_at)

    def _is_analysis_completed(self, analysis: Analysis) -> bool:
        return self.analysis_client.is_latest_analysis_completed(case_id=analysis.case.internal_id)

    @staticmethod
    def _has_uploaded_started(analysis: Analysis) -> bool:
        return bool(analysis.upload_started_at)



"""Tests for chanjo coverage API."""

from pathlib import Path

from cg.apps.coverage.api import ChanjoAPI
from cg.utils.commands import Process


def test_chanjo_api_init(chanjo_config: dict[str, dict[str, str]]):
    """Test __init__."""
    # GIVEN a config dict

    # WHEN instantiating a chanjo API
    api = ChanjoAPI(chanjo_config)

    # THEN attributes chanjo_config and chanjo_binary are set
    assert api.chanjo_config == chanjo_config["chanjo"]["config_path"]
    assert api.chanjo_binary == chanjo_config["chanjo"]["binary_path"]


def test_chanjo_api_upload(
    bed_file: Path,
    case_id: str,
    chanjo_config: dict[str, dict[str, str]],
    family_name: str,
    mocker,
    sample_id: str,
    sample_name: str,
):
    """Test upload method."""
    # GIVEN sample_id, sample_name, group_id, group_name, and a bed_file

    # WHEN uploading a sample with the api, using a mocked Process.run_command method
    api = ChanjoAPI(chanjo_config)
    mocked_run_command = mocker.patch.object(Process, "run_command")
    api.upload(
        sample_id=sample_id,
        sample_name=sample_name,
        group_id=case_id,
        group_name=family_name,
        bed_file=bed_file.as_posix(),
    )

    # THEN run_command should be called with the list
    mocked_run_command.assert_called_once_with(
        parameters=[
            "load",
            "--sample",
            sample_id,
            "--name",
            sample_name,
            "--group",
            case_id,
            "--group-name",
            family_name,
            "--threshold",
            "10",
            bed_file.as_posix(),
        ]
    )


def test_chanjo_api_sample_existing(
    chanjo_config: dict[str, dict[str, str]],
    mocker,
    mock_process,
    sample_id: str,
):
    """Test sample method."""

    # GIVEN a sample_id

    # WHEN fetching an existing sample from the api, with a mocked stdout
    mocked_stdout = '[{"id": "%s"}]' % sample_id
    mocked_stderr = ""
    MockedProcess = mock_process(result_stderr=mocked_stderr, result_stdout=mocked_stdout)
    mocked_process = mocker.patch("cg.apps.coverage.api.Process")
    mocked_process.return_value = MockedProcess(
        binary=chanjo_config["chanjo"]["binary_path"],
        config=chanjo_config["chanjo"]["config_path"],
    )
    api = ChanjoAPI(chanjo_config)
    sample = api.sample(sample_id=sample_id)

    # THEN sample should be a dictionary with key id = sample_id
    assert sample["id"] == sample_id


def test_chanjo_api_sample_non_existing(
    chanjo_config: dict[str, dict[str, str]], mocker, mock_process, sample_id: str
):
    """Test sample method."""

    # GIVEN a sample_id

    # WHEN fetching a non existing sample from the api, with a mocked stdout
    mocked_stdout = "[]"
    mocked_stderr = ""
    MockedProcess = mock_process(result_stderr=mocked_stderr, result_stdout=mocked_stdout)
    mocked_process = mocker.patch("cg.apps.coverage.api.Process")
    mocked_process.return_value = MockedProcess(
        binary=chanjo_config["chanjo"]["binary_path"],
        config=chanjo_config["chanjo"]["config_path"],
    )
    api = ChanjoAPI(chanjo_config)
    sample = api.sample(sample_id=sample_id)

    # THEN None should have been returned
    assert sample is None


def test_chanjo_api_delete_sample(chanjo_config: dict[str, dict[str, str]], mocker, sample_id: str):
    """Test delete method."""
    # GIVEN a sample_id

    # WHEN deleting a sample with the api and a mocked Process.run_command method
    mocked_run_command = mocker.patch.object(Process, "run_command")
    api = ChanjoAPI(chanjo_config)
    api.delete_sample(sample_id=sample_id)

    # THEN run_command should be called once with list ["db", "remove", sample_id]
    mocked_run_command.assert_called_once_with(parameters=["db", "remove", sample_id])


def test_chanjo_api_omim_coverage(
    chanjo_config: dict[str, dict[str, str]],
    chanjo_mean_completeness: int,
    chanjo_mean_coverage: int,
    mocker,
    mock_process,
    sample_id: str,
):
    """Test omim_coverage method."""
    # GIVEN a sample_id

    # WHEN using the omim_coverage method with a mocked stdout
    mocked_stdout = '{"%s": {"mean_coverage": %f, "mean_completeness": %f}}' % (
        sample_id,
        chanjo_mean_coverage,
        chanjo_mean_completeness,
    )
    mocked_stderr = ""
    MockedProcess = mock_process(result_stderr=mocked_stderr, result_stdout=mocked_stdout)
    mocked_process = mocker.patch("cg.apps.coverage.api.Process")
    mocked_process.return_value = MockedProcess(
        binary=chanjo_config["chanjo"]["binary_path"],
        config=chanjo_config["chanjo"]["config_path"],
    )
    api = ChanjoAPI(chanjo_config)
    samples = api.omim_coverage(samples=[{"id": sample_id}])

    # THEN this should return a dictionary with mean coverage and mean_completeness
    # for each sample
    assert samples[sample_id]["mean_coverage"] == chanjo_mean_coverage
    assert samples[sample_id]["mean_completeness"] == chanjo_mean_completeness


def test_chanjo_api_coverage(
    chanjo_config: dict[str, dict[str, str]],
    chanjo_mean_completeness: int,
    chanjo_mean_coverage: int,
    mocker,
    mock_process,
    sample_id: str,
):
    """Test coverage method."""
    # GIVEN a sample_id

    # WHEN using the coverage method with a mocked stdout
    mocked_stdout = '{"%s": {"mean_coverage": %f, "mean_completeness": %f}}' % (
        sample_id,
        chanjo_mean_coverage,
        chanjo_mean_completeness,
    )
    mocked_stderr = ""
    MockedProcess = mock_process(result_stderr=mocked_stderr, result_stdout=mocked_stdout)
    mocked_process = mocker.patch("cg.apps.coverage.api.Process")
    mocked_process.return_value = MockedProcess(
        binary=chanjo_config["chanjo"]["binary_path"],
        config=chanjo_config["chanjo"]["config_path"],
    )
    api = ChanjoAPI(chanjo_config)
    samples = api.sample_coverage(sample_id=sample_id, panel_genes=["123"])

    # THEN this should return a dictionary with mean coverage and mean_completeness
    # the sample
    assert samples["mean_coverage"] == chanjo_mean_coverage
    assert samples["mean_completeness"] == chanjo_mean_completeness



"""Tests for rsync API"""

import logging
import shutil
from pathlib import Path

import pytest

from cg.constants import Workflow
from cg.constants.priority import SlurmAccount, SlurmQos
from cg.exc import CgError
from cg.services.deliver_files.delivery_rsync_service.delivery_rsync_service import (
    DeliveryRsyncService,
)
from cg.store.models import Case
from cg.store.store import Store
from tests.store.conftest import case_obj


def test_get_source_and_destination_paths(
    mutant_case: Case, delivery_rsync_service: DeliveryRsyncService, ticket_id: str, mocker
):
    """Test generating the source path before rsync."""

    # GIVEN a valid Sars-cov-2 case
    case = mutant_case

    # GIVEN file exists
    mocker.patch.object(DeliveryRsyncService, "get_all_cases_from_ticket")
    DeliveryRsyncService.get_all_cases_from_ticket.return_value = [case]

    # WHEN the source path is created
    source_and_destination_paths = delivery_rsync_service.get_source_and_destination_paths(
        ticket=ticket_id, customer_internal_id=case.customer.internal_id
    )

    # THEN the source path ends with a customer id, followed by "inbox" and a ticket_id id
    assert (
        source_and_destination_paths["delivery_source_path"]
        .as_posix()
        .endswith(f"/cust000/inbox/{ticket_id}")
    )
    # THEN the destination path is in the format server.name.se:/path/cust_id/path/ticket_id/
    assert (
        source_and_destination_paths["rsync_destination_path"].as_posix()
        == "server.name.se:/some/cust000/inbox"
    )


def test_set_log_dir(delivery_rsync_service: DeliveryRsyncService, ticket_id: str, caplog):
    """Test function to set log dir for path."""

    caplog.set_level(logging.INFO)

    # GIVEN an DeliveryRsyncService, with its base path as its log dir
    base_path: Path = delivery_rsync_service.log_dir

    # WHEN setting the log directory
    delivery_rsync_service.set_log_dir(folder_prefix=ticket_id)

    # THEN the log dir should set to a new path, different from the base path
    assert base_path.as_posix() != delivery_rsync_service.log_dir.as_posix()
    assert "Setting log dir to:" in caplog.text


def test_make_log_dir(delivery_rsync_service: DeliveryRsyncService, ticket_id: str, caplog):
    """Test generating the directory for logging."""
    caplog.set_level(logging.INFO)

    # WHEN the log directory is created
    delivery_rsync_service.set_log_dir(folder_prefix=ticket_id)
    delivery_rsync_service.create_log_dir(dry_run=True)

    # THEN the path is not created since it is a dry run
    assert "Would have created path" in caplog.text

    # THEN the created path is
    assert str(delivery_rsync_service.log_dir).startswith(f"/another/path/{ticket_id}")


def test_run_rsync_on_slurm(
    microsalt_case: Case,
    delivery_rsync_service: DeliveryRsyncService,
    ticket_id: str,
    caplog,
    mocker,
    helpers,
):
    """Test for running rsync using SLURM."""
    caplog.set_level(logging.INFO)

    # GIVEN a valid microsalt case
    case: Case = microsalt_case

    # GIVEN paths needed to run rsync
    mocker.patch.object(DeliveryRsyncService, "get_source_and_destination_paths")
    DeliveryRsyncService.get_source_and_destination_paths.return_value = {
        "delivery_source_path": Path("/path/to/source"),
        "rsync_destination_path": Path("/path/to/destination"),
    }

    mocker.patch.object(DeliveryRsyncService, "get_all_cases_from_ticket")
    DeliveryRsyncService.get_all_cases_from_ticket.return_value = [case]

    # WHEN the destination path is created
    sbatch_number: int = delivery_rsync_service.run_rsync_for_ticket(ticket=ticket_id, dry_run=True)

    # THEN check that SARS-COV-2 analysis is not delivered
    assert "Delivering report for SARS-COV-2 analysis" not in caplog.text

    # THEN check that an integer was returned as sbatch number
    assert isinstance(sbatch_number, int)


def test_run_rsync_on_slurm_no_cases(
    delivery_rsync_service: DeliveryRsyncService, ticket_id: str, caplog, mocker, helpers
):
    """Test for running rsync using SLURM when there are no cases on the ticket."""
    caplog.set_level(logging.INFO)

    # GIVEN ticket without any cases
    mocker.patch.object(DeliveryRsyncService, "get_all_cases_from_ticket")
    DeliveryRsyncService.get_all_cases_from_ticket.return_value = None

    # WHEN the job is submitted

    # THEN an error is raised
    with pytest.raises(CgError):
        delivery_rsync_service.run_rsync_for_ticket(ticket=ticket_id, dry_run=True)


def test_concatenate_rsync_commands(
    analysis_family: dict,
    analysis_store_trio,
    case: Case,
    customer_id,
    folders_to_deliver: set[Path],
    project_dir,
    delivery_rsync_service: DeliveryRsyncService,
    ticket_id: str,
):
    """Tests the function to concatenate rsync commands for transferring multiple files."""
    # GIVEN a list with a case and a sample name

    source_and_destination_paths = {
        "delivery_source_path": Path(project_dir, customer_id, ticket_id),
        "rsync_destination_path": Path(project_dir, customer_id),
    }
    # WHEN then commands are generated
    command: str = delivery_rsync_service.concatenate_rsync_commands(
        folder_list=folders_to_deliver,
        source_and_destination_paths=source_and_destination_paths,
        ticket=ticket_id,
        case=case,
    )
    # THEN the correct folder should be added to the source path
    assert (
        " ".join(
            [
                str(source_and_destination_paths["delivery_source_path"] / analysis_family["name"]),
                str(source_and_destination_paths["delivery_source_path"]),
            ]
        )
        in command
    )
    assert (
        " ".join(
            [
                str(
                    source_and_destination_paths["delivery_source_path"]
                    / analysis_family["samples"][0]["name"]
                ),
                str(source_and_destination_paths["delivery_source_path"]),
            ]
        )
        in command
    )


def test_concatenate_rsync_commands_mutant(
    analysis_family: dict,
    analysis_store_trio,
    case: Case,
    customer_id,
    folders_to_deliver: set[Path],
    mocker,
    project_dir,
    delivery_rsync_service: DeliveryRsyncService,
    ticket_id: str,
):
    """Tests the function to concatenate Rsync commands for transferring multiple files."""
    # GIVEN a list with a Mutant case and a sample name and a Mutant report file
    case.data_analysis = Workflow.MUTANT
    source_and_destination_paths = {
        "delivery_source_path": Path(project_dir, customer_id, ticket_id),
        "rsync_destination_path": Path(project_dir, customer_id),
    }
    report_path = Path(project_dir, customer_id, ticket_id, "a_report_file")
    covid_destination_path = Path(project_dir, "destination")
    delivery_rsync_service.covid_destination_path = covid_destination_path

    # WHEN then commands are generated
    mocker.patch.object(DeliveryRsyncService, "format_covid_report_path", return_value=report_path)
    mocker.patch.object(
        DeliveryRsyncService, "format_covid_destination_path", return_value=covid_destination_path
    )
    command: str = delivery_rsync_service.concatenate_rsync_commands(
        folder_list=folders_to_deliver,
        source_and_destination_paths=source_and_destination_paths,
        ticket=ticket_id,
        case=case,
    )

    # THEN the correct folder should be added to the source path
    assert report_path.name in command
    assert covid_destination_path.as_posix() in command


def test_slurm_rsync_single_case(
    all_samples_in_inbox: Path,
    case: Case,
    destination_path: Path,
    delivery_rsync_service: DeliveryRsyncService,
    caplog,
    mocker,
    ticket_id: str,
    folders_to_deliver: set[Path],
):
    """Test for running rsync on a single case using SLURM."""
    caplog.set_level(logging.INFO)

    # GIVEN paths needed to run rsync
    mocker.patch.object(DeliveryRsyncService, "get_source_and_destination_paths")
    DeliveryRsyncService.get_source_and_destination_paths.return_value = {
        "delivery_source_path": all_samples_in_inbox,
        "rsync_destination_path": destination_path,
    }

    mocker.patch.object(Store, "get_latest_ticket_from_case")
    Store.get_latest_ticket_from_case.return_value = ticket_id

    # WHEN the destination path is created
    sbatch_number: int
    is_complete_delivery: bool
    sbatch_number: int = delivery_rsync_service.run_rsync_for_case(
        case=case,
        dry_run=True,
        folders_to_deliver=folders_to_deliver,
    )

    # THEN check that an integer was returned as sbatch number and the delivery should be complete
    assert isinstance(sbatch_number, int)


def test_slurm_rsync_single_case_missing_file(
    all_samples_in_inbox: Path,
    case: Case,
    destination_path: Path,
    delivery_rsync_service: DeliveryRsyncService,
    caplog,
    mocker,
    ticket_id: str,
    folders_to_deliver: set[Path],
):
    """Test for running rsync on a single case with a missing file using SLURM."""
    caplog.set_level(logging.INFO)

    # GIVEN a valid mip case and sample folder missing
    shutil.rmtree(Path(all_samples_in_inbox, case.links[0].sample.name))

    # GIVEN paths needed to run rsync
    mocker.patch.object(DeliveryRsyncService, "get_source_and_destination_paths")
    DeliveryRsyncService.get_source_and_destination_paths.return_value = {
        "delivery_source_path": all_samples_in_inbox,
        "rsync_destination_path": destination_path,
    }

    mocker.patch.object(Store, "get_latest_ticket_from_case")
    Store.get_latest_ticket_from_case.return_value = ticket_id

    # WHEN the destination path is created
    sbatch_number: int

    sbatch_number: int = delivery_rsync_service.run_rsync_for_case(
        case=case, dry_run=True, folders_to_deliver=folders_to_deliver
    )

    # THEN check that an integer was returned as sbatch number
    assert isinstance(sbatch_number, int)


def test_slurm_quality_of_service_production(delivery_rsync_service: DeliveryRsyncService):
    # GIVEN an DeliveryRsyncService instance

    # WHEN the account of the api is set to production
    delivery_rsync_service.account = SlurmAccount.PRODUCTION

    # THEN the quality of service should be set to HIGH
    assert delivery_rsync_service.slurm_quality_of_service == SlurmQos.HIGH


def test_slurm_quality_of_service_other(delivery_rsync_service: DeliveryRsyncService):
    # GIVEN an DeliveryRsyncService instance

    # WHEN the account of the api is set to development
    delivery_rsync_service.account = SlurmAccount.DEVELOPMENT

    # THEN the quality of service should be set to LOW
    assert delivery_rsync_service.slurm_quality_of_service == SlurmQos.LOW



from pydantic_settings import BaseSettings


class AppConfig(BaseSettings):
    """Default values are overridden by environment variable at run-time."""

    gunicorn_workers: int = 4
    gunicorn_threads: int = 4
    gunicorn_bind: str = "0.0.0.0:8000"
    gunicorn_timeout: int = 400
    cg_sql_database_uri: str = "sqlite:///"
    cg_secret_key: str = "thisIsNotASafeKey"
    invoice_max_price: int = 750_000
    lims_host: str = "lims_host"
    lims_username: str = "username"
    lims_password: str = "password"
    support_system_email: str = "support@mail.com"
    email_uri: str = "smtp://localhost"
    google_oauth_client_id: str = "client_id"
    google_oauth_client_secret: str = "client_secret"
    trailblazer_host: str = "trailblazer_host"
    trailblazer_service_account: str = "service_account"
    trailblazer_service_account_auth_file: str = "auth_file.json"
    freshdesk_url: str = "https://company.freshdesk.com"
    freshdesk_api_key: str = "freshdesk_api_key"
    freshdesk_order_email_id: int = 10
    freshdesk_environment: str = "Stage"


app_config = AppConfig()



"""Delivery report helpers."""

import logging
from datetime import datetime

import click

from cg.constants import (
    REPORT_SUPPORTED_DATA_DELIVERY,
    REPORT_SUPPORTED_WORKFLOW,
    Workflow,
)
from cg.meta.report.balsamic import BalsamicReportAPI
from cg.meta.report.balsamic_qc import BalsamicQCReportAPI
from cg.meta.report.balsamic_umi import BalsamicUmiReportAPI
from cg.meta.report.mip_dna import MipDNAReportAPI
from cg.meta.report.raredisease import RarediseaseReportAPI
from cg.meta.report.report_api import ReportAPI
from cg.meta.report.rnafusion import RnafusionReportAPI
from cg.meta.report.taxprofiler import TaxprofilerReportAPI
from cg.meta.report.tomte import TomteReportAPI
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.meta.workflow.balsamic_qc import BalsamicQCAnalysisAPI
from cg.meta.workflow.balsamic_umi import BalsamicUmiAnalysisAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.meta.workflow.raredisease import RarediseaseAnalysisAPI
from cg.meta.workflow.rnafusion import RnafusionAnalysisAPI
from cg.meta.workflow.taxprofiler import TaxprofilerAnalysisAPI
from cg.meta.workflow.tomte import TomteAnalysisAPI
from cg.store.models import Case

LOG = logging.getLogger(__name__)


def get_report_case(context: click.Context, case_id: str) -> Case:
    """Extracts a case object for delivery report generation."""
    # Default report API (MIP DNA report API)
    report_api: ReportAPI = (
        context.obj.meta_apis.get("report_api")
        if context.obj.meta_apis.get("report_api")
        else MipDNAReportAPI(config=context.obj, analysis_api=MipDNAAnalysisAPI(config=context.obj))
    )
    case: Case = report_api.status_db.get_case_by_internal_id(internal_id=case_id)
    # Missing or not valid internal case ID
    if not case_id or not case:
        LOG.warning("Invalid case ID. Retrieving available cases.")
        workflow: Workflow = (
            report_api.analysis_api.workflow if context.obj.meta_apis.get("report_api") else None
        )
        cases_without_delivery_report: list[Case] = (
            report_api.get_cases_without_delivery_report(workflow=workflow)
            if not context.obj.meta_apis.get("upload_api")
            else report_api.get_cases_without_uploaded_delivery_report(workflow=workflow)
        )
        if not cases_without_delivery_report:
            click.echo(
                click.style(
                    "There are no valid cases to perform delivery report actions", fg="green"
                )
            )
        else:
            LOG.error("Provide one of the following case IDs:")
            for case in cases_without_delivery_report:
                click.echo(f"{case.internal_id} ({case.data_analysis})")
        raise click.Abort
    if case.data_analysis not in REPORT_SUPPORTED_WORKFLOW:
        LOG.error(
            f"The {case.data_analysis} workflow does not support delivery reports (case: {case.internal_id})"
        )
        raise click.Abort
    if case.data_delivery not in REPORT_SUPPORTED_DATA_DELIVERY:
        LOG.error(
            f"The {case.data_delivery} data delivery does not support delivery reports (case: {case.internal_id})"
        )
        raise click.Abort
    return case


def get_report_api(context: click.Context, case: Case) -> ReportAPI:
    """Returns a report API to be used for the delivery report generation."""
    if context.obj.meta_apis.get("report_api"):
        return context.obj.meta_apis.get("report_api")
    return get_report_api_workflow(context, case.data_analysis)


def get_report_api_workflow(context: click.Context, workflow: Workflow) -> ReportAPI:
    """Return the report API given a specific workflow."""
    # Default report API workflow: MIP-DNA
    workflow: Workflow = workflow if workflow else Workflow.MIP_DNA
    dispatch_report_api: dict[Workflow, ReportAPI] = {
        Workflow.BALSAMIC: BalsamicReportAPI(
            config=context.obj, analysis_api=BalsamicAnalysisAPI(config=context.obj)
        ),
        Workflow.BALSAMIC_QC: BalsamicQCReportAPI(
            config=context.obj, analysis_api=BalsamicQCAnalysisAPI(config=context.obj)
        ),
        Workflow.BALSAMIC_UMI: BalsamicUmiReportAPI(
            config=context.obj, analysis_api=BalsamicUmiAnalysisAPI(config=context.obj)
        ),
        Workflow.MIP_DNA: MipDNAReportAPI(
            config=context.obj, analysis_api=MipDNAAnalysisAPI(config=context.obj)
        ),
        Workflow.RAREDISEASE: RarediseaseReportAPI(
            config=context.obj, analysis_api=RarediseaseAnalysisAPI(config=context.obj)
        ),
        Workflow.RNAFUSION: RnafusionReportAPI(
            config=context.obj, analysis_api=RnafusionAnalysisAPI(config=context.obj)
        ),
        Workflow.TAXPROFILER: TaxprofilerReportAPI(
            config=context.obj, analysis_api=TaxprofilerAnalysisAPI(config=context.obj)
        ),
        Workflow.TOMTE: TomteReportAPI(
            config=context.obj, analysis_api=TomteAnalysisAPI(config=context.obj)
        ),
    }
    return dispatch_report_api.get(workflow)


def get_report_analysis_started(
    case: Case, report_api: ReportAPI, analysis_started_at: str | None
) -> datetime:
    """Resolves and returns a valid analysis date."""
    if not analysis_started_at:
        analysis_started_at: datetime = (
            report_api.status_db.get_case_by_internal_id(internal_id=case.internal_id)
            .analyses[0]
            .started_at
        )
    # If there is no analysis for the provided date
    if not report_api.status_db.get_analysis_by_case_entry_id_and_started_at(
        case_entry_id=case.id, started_at_date=analysis_started_at
    ):
        LOG.error(f"There is no analysis started at {analysis_started_at}")
        raise click.Abort
    LOG.info(f"Using analysis started at: {analysis_started_at}")
    return analysis_started_at



from pathlib import Path

import pytest

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.delivery import INBOX_NAME
from cg.constants.housekeeper_tags import (
    HK_DELIVERY_REPORT_TAG,
    AlignmentFileTag,
    SequencingFileTag,
)
from cg.services.deliver_files.delivery_file_fetcher_service.models import (
    DeliveryFiles,
    SampleFile,
    CaseFile,
    DeliveryMetaData,
)
from cg.store.models import Case
from cg.store.store import Store


@pytest.fixture
def expected_fastq_delivery_files(
    delivery_housekeeper_api: HousekeeperAPI,
    case_id: str,
    sample_id: str,
    sample_name: str,
    another_sample_id: str,
    another_sample_name: str,
    delivery_store_microsalt: Store,
) -> DeliveryFiles:
    """Return the expected fastq delivery files."""
    sample_info: list[tuple[str, str]] = [
        (sample_id, sample_name),
        (another_sample_id, another_sample_name),
    ]
    sample_files: list[SampleFile] = [
        SampleFile(
            case_id=case_id,
            sample_id=sample[0],
            sample_name=sample[1],
            file_path=delivery_housekeeper_api.get_files_from_latest_version(
                bundle_name=sample[0], tags=[SequencingFileTag.FASTQ]
            )[0].full_path,
        )
        for sample in sample_info
    ]
    case: Case = delivery_store_microsalt.get_case_by_internal_id(case_id)
    delivery_meta_data = DeliveryMetaData(
        customer_internal_id=case.customer.internal_id, ticket_id=case.latest_ticket
    )
    return DeliveryFiles(
        delivery_data=delivery_meta_data, case_files=None, sample_files=sample_files
    )


@pytest.fixture
def expected_analysis_delivery_files(
    delivery_housekeeper_api: HousekeeperAPI,
    case_id: str,
    case_name: str,
    sample_id: str,
    sample_name: str,
    another_sample_id: str,
    another_sample_name: str,
    delivery_store_balsamic: Store,
) -> DeliveryFiles:
    """Return the expected analysis delivery files."""
    sample_info: list[tuple[str, str]] = [
        (sample_id, sample_name),
        (another_sample_id, another_sample_name),
    ]
    sample_files: list[SampleFile] = []
    for sample in sample_info:
        sample_files.extend(
            [
                SampleFile(
                    case_id=case_id,
                    sample_id=sample[0],
                    sample_name=sample[1],
                    file_path=file.full_path,
                )
                for file in delivery_housekeeper_api.get_files_from_latest_version(
                    bundle_name=case_id, tags=[AlignmentFileTag.CRAM, sample[0]]
                )
            ]
        )
    case_files: list[CaseFile] = [
        CaseFile(
            case_id=case_id,
            case_name=case_name,
            file_path=delivery_housekeeper_api.get_files_from_latest_version(
                bundle_name=case_id, tags=[HK_DELIVERY_REPORT_TAG]
            )[0].full_path,
        )
    ]
    case: Case = delivery_store_balsamic.get_case_by_internal_id(case_id)
    delivery_meta_data = DeliveryMetaData(
        customer_internal_id=case.customer.internal_id, ticket_id=case.latest_ticket
    )
    return DeliveryFiles(
        delivery_data=delivery_meta_data, case_files=case_files, sample_files=sample_files
    )


@pytest.fixture
def expected_moved_fastq_delivery_files(
    expected_fastq_delivery_files: DeliveryFiles, tmp_path: Path
) -> DeliveryFiles:
    """Return the moved FASTQ delivery files."""
    delivery_files = DeliveryFiles(**expected_fastq_delivery_files.model_dump())
    inbox_dir_path = Path(
        tmp_path,
        delivery_files.delivery_data.customer_internal_id,
        INBOX_NAME,
        delivery_files.delivery_data.ticket_id,
    )
    new_sample_files: list[SampleFile] = swap_file_paths_with_inbox_paths(
        file_models=delivery_files.sample_files, inbox_dir_path=inbox_dir_path
    )
    return DeliveryFiles(
        delivery_data=expected_fastq_delivery_files.delivery_data,
        case_files=None,
        sample_files=new_sample_files,
    )


@pytest.fixture
def expected_moved_analysis_delivery_files(
    expected_analysis_delivery_files: DeliveryFiles, tmp_path: Path
) -> DeliveryFiles:
    """Return the moved analysis delivery files."""
    delivery_files = DeliveryFiles(**expected_analysis_delivery_files.model_dump())
    inbox_dir_path = Path(
        tmp_path,
        delivery_files.delivery_data.customer_internal_id,
        INBOX_NAME,
        delivery_files.delivery_data.ticket_id,
    )
    new_case_files: list[CaseFile] = swap_file_paths_with_inbox_paths(
        file_models=delivery_files.case_files, inbox_dir_path=inbox_dir_path
    )
    new_sample_files: list[SampleFile] = swap_file_paths_with_inbox_paths(
        file_models=delivery_files.sample_files, inbox_dir_path=inbox_dir_path
    )
    return DeliveryFiles(
        delivery_data=delivery_files.delivery_data,
        case_files=new_case_files,
        sample_files=new_sample_files,
    )


@pytest.fixture
def expected_moved_analysis_sample_delivery_files(
    expected_moved_analysis_delivery_files: DeliveryFiles,
) -> list[SampleFile]:
    return expected_moved_analysis_delivery_files.sample_files


@pytest.fixture
def expected_moved_analysis_case_delivery_files(
    expected_moved_analysis_delivery_files: DeliveryFiles,
) -> list[CaseFile]:
    return expected_moved_analysis_delivery_files.case_files


@pytest.fixture
def fastq_concatenation_sample_files(tmp_path: Path) -> list[SampleFile]:
    some_ticket: str = "some_ticket"
    fastq_paths: list[Path] = [
        Path(tmp_path, some_ticket, "S1_1_R1_1.fastq.gz"),
        Path(tmp_path, some_ticket, "S1_2_R1_1.fastq.gz"),
        Path(tmp_path, some_ticket, "S1_1_R2_1.fastq.gz"),
        Path(tmp_path, some_ticket, "S1_2_R2_1.fastq.gz"),
    ]
    return [
        SampleFile(
            sample_id="S1",
            case_id="Case1",
            sample_name="Sample1",
            file_path=fastq_path,
        )
        for fastq_path in fastq_paths
    ]


def swap_file_paths_with_inbox_paths(
    file_models: list[CaseFile | SampleFile], inbox_dir_path: Path
) -> list[CaseFile | SampleFile]:
    """Swap the file paths with the inbox paths."""
    new_file_models: list[SampleFile | CaseFile] = []
    for file_model in file_models:
        new_file_model: SampleFile = file_model
        new_file_model.file_path = Path(inbox_dir_path, file_model.file_path.name)
        new_file_models.append(new_file_model)
    return new_file_models



"""Tests for the PacBioRunDataGenerator."""

from pathlib import Path

import pytest

from cg.services.run_devices.exc import PostProcessingRunDataGeneratorError
from cg.services.run_devices.pacbio.run_data_generator.pacbio_run_data_generator import (
    PacBioRunDataGenerator,
)
from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData


def test_get_run_data(
    pac_bio_run_data_generator: PacBioRunDataGenerator,
    pac_bio_runs_dir: Path,
    pac_bio_test_run_name: str,
    pac_bio_smrt_cell_name: str,
    expected_pac_bio_run_data: PacBioRunData,
):
    # GIVEN a run directory, a run name and a SMRT cell name
    run_name: str = "/".join([pac_bio_test_run_name, pac_bio_smrt_cell_name])

    # WHEN Generating run data
    run_data: PacBioRunData = pac_bio_run_data_generator.get_run_data(
        run_name=run_name, sequencing_dir=pac_bio_runs_dir.as_posix()
    )

    # THEN the correct run data are returned
    assert run_data == expected_pac_bio_run_data


@pytest.mark.parametrize("wrong_run_name", ["rimproper_name", "d_improper_name "])
def test_get_run_data_improper_name(
    pac_bio_run_data_generator: PacBioRunDataGenerator,
    pac_bio_runs_dir: Path,
    wrong_run_name: str,
):
    # GIVEN a PacBioRunDataGenerator and a wrong run name

    # WHEN Generating run data with the wrong run name

    # THEN an PostProcessingRunDataGeneratorError is raised
    with pytest.raises(PostProcessingRunDataGeneratorError):
        pac_bio_run_data_generator.get_run_data(
            run_name=wrong_run_name, sequencing_dir=pac_bio_runs_dir.as_posix()
        )



"""Tests CLI common methods to assess workflow base command for NF analyses."""

import pytest
from click.testing import CliRunner

from cg.cli.workflow.base import workflow as workflow_cli
from cg.constants import EXIT_SUCCESS, Workflow
from cg.constants.nextflow import NEXTFLOW_WORKFLOWS


@pytest.mark.parametrize(
    "workflow",
    NEXTFLOW_WORKFLOWS + [Workflow.JASEN],
)
def test_workflow_no_args(cli_runner: CliRunner, workflow: Workflow, request):
    """Test to see that workflow is added and prints help when no subcommand is specified."""
    context = request.getfixturevalue(f"{workflow}_context")
    # GIVEN no arguments or options besides the command call

    # WHEN running command
    result = cli_runner.invoke(workflow_cli, [workflow], obj=context)

    # THEN command runs successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN help should be printed
    assert "help" in result.output



from datetime import datetime, timedelta
from pathlib import Path
from unittest import mock
from urllib.parse import urljoin

import pytest
from requests import Response

from cg.constants.constants import APIMethods, FileFormat
from cg.exc import DdnDataflowAuthenticationError
from cg.io.controller import APIRequest, WriteStream
from cg.meta.archive.ddn.constants import (
    DESTINATION_ATTRIBUTE,
    OSTYPE,
    ROOT_TO_TRIM,
    SOURCE_ATTRIBUTE,
    DataflowEndpoints,
)
from cg.meta.archive.ddn.ddn_data_flow_client import DDNDataFlowClient
from cg.meta.archive.ddn.models import MiriaObject, TransferPayload
from cg.meta.archive.ddn.utils import get_metadata
from cg.meta.archive.models import FileAndSample
from cg.models.cg_config import DataFlowConfig
from cg.store.store import Store

FUNCTION_TO_MOCK = "cg.meta.archive.ddn.ddn_data_flow_client.APIRequest.api_request_from_content"


def test_correct_source_root(
    local_directory: Path, miria_file_archive: MiriaObject, trimmed_local_directory: Path
):
    """Tests the method for trimming the source directory."""

    # GIVEN a MiriaObject with a source path and a destination path

    # WHEN trimming the path of the source attribute
    miria_file_archive.trim_path(attribute_to_trim=SOURCE_ATTRIBUTE)

    # THEN the source path should be the local directory minus the /home part
    assert miria_file_archive.source == trimmed_local_directory.as_posix()


def test_correct_destination_root(
    local_directory: Path, miria_file_archive: MiriaObject, trimmed_local_directory: Path
):
    """Tests the method for trimming the destination directory."""

    # GIVEN a MiriaObject with a source path and a destination path
    miria_file_archive.destination = local_directory

    # WHEN trimming the path of the destination attribute
    miria_file_archive.trim_path(attribute_to_trim=DESTINATION_ATTRIBUTE)

    # THEN the destination path should be the local directory minus the /home part
    assert miria_file_archive.destination == trimmed_local_directory.as_posix()


def test_add_repositories(
    ddn_dataflow_config, local_directory, remote_path, miria_file_archive: MiriaObject
):
    """Tests the method for adding the repositories to the source and destination paths."""

    # GIVEN a MiriaObject object

    # WHEN adding the repositories
    miria_file_archive.add_repositories(
        source_prefix=ddn_dataflow_config.local_storage,
        destination_prefix=ddn_dataflow_config.archive_repository,
    )

    # THEN the repositories should be prepended to the paths
    assert miria_file_archive.source == ddn_dataflow_config.local_storage + str(local_directory)
    assert miria_file_archive.destination == ddn_dataflow_config.archive_repository + str(
        remote_path
    )


def test_transfer_payload_model_dump(transfer_payload: TransferPayload):
    """Tests that the dict structure returned by TransferPayload.dict() is compatible with the
    Dataflow API."""

    # GIVEN a TransferPayload object with two MiriaObject objects

    # WHEN obtaining the dict representation
    dict_representation: dict = transfer_payload.model_dump(by_alias=True)

    # THEN the following fields should exist
    assert dict_representation.get("pathInfo", None)
    assert dict_representation.get("osType", None)
    assert isinstance(dict_representation.get("metadataList", None), list)

    # THEN the pathInfo fields should contain source and destination fields
    assert dict_representation.get("pathInfo")[0].get("source", None)
    assert dict_representation.get("pathInfo")[0].get("destination", None)


def test_ddn_dataflow_client_initialization(
    future_date: datetime,
    local_storage_repository: str,
    remote_storage_repository: str,
    ok_response: Response,
):
    """Tests the initialization of the DDNDataFlowClient object with given an ok-response."""

    # GIVEN a valid DDNConfig object
    valid_config = DataFlowConfig(
        database_name="test_database",
        user="test_user",
        password="test_password",
        url="https://test-url.com",
        archive_repository=remote_storage_repository,
        local_storage=local_storage_repository,
    )

    # GIVEN a mock response with a 200 OK status code and valid JSON content
    ok_response._content = WriteStream.write_stream_from_content(
        content={
            "refresh": "test_refresh_token",
            "access": "test_access_token",
            "expire": future_date.timestamp(),
        },
        file_format=FileFormat.JSON,
    ).encode()

    # WHEN initializing the DDNDataFlowClient class with the valid DDNConfig object
    with mock.patch(
        FUNCTION_TO_MOCK,
        return_value=ok_response,
    ):
        ddn_dataflow_client = DDNDataFlowClient(config=valid_config)

    # THEN the DDNDataFlowClient object should be created successfully
    assert isinstance(ddn_dataflow_client, DDNDataFlowClient)


def test_set_auth_tokens(ddn_dataflow_config: DataFlowConfig, ok_response: Response):
    """Tests the functions setting the auth- and refresh token as well as the expiration date."""

    # GIVEN a valid DDNConfig object
    # GIVEN a mock response with the auth and refresh tokens
    ok_response._content = (
        b'{"access": "test_auth_token", "refresh": "test_refresh_token", "expire": 1677649423}'
    )

    # WHEN initializing the DDNDataFlowClient class with the valid DDNConfig object
    with mock.patch(
        FUNCTION_TO_MOCK,
        return_value=ok_response,
    ):
        ddn_dataflow_client = DDNDataFlowClient(config=ddn_dataflow_config)

    # THEN the auth and refresh tokens should be set correctly
    assert ddn_dataflow_client.auth_token == "test_auth_token"
    assert ddn_dataflow_client.refresh_token == "test_refresh_token"


def test_ddn_dataflow_client_initialization_invalid_credentials(
    ddn_dataflow_config: DataFlowConfig, unauthorized_response: Response
):
    """Tests initialization of a DDNDataFlowClient when an error is returned."""

    # GIVEN a valid DDNConfig object with invalid credentials
    ddn_dataflow_config.password = "Wrong_password"

    # GIVEN a mock response with a 401 Unauthorized status code
    unauthorized_response._content = b'{"detail": "Invalid credentials"}'

    # WHEN initializing the DDNDataFlowClient class with the invalid credentials
    with mock.patch(
        "cg.meta.archive.ddn.ddn_data_flow_client.APIRequest.api_request_from_content",
        return_value=unauthorized_response,
    ):
        # THEN an exception should be raised
        with pytest.raises(DdnDataflowAuthenticationError):
            DDNDataFlowClient(config=ddn_dataflow_config)


def test_transfer_payload_correct_source_root(transfer_payload: TransferPayload):
    """Tests trimming all source paths in the TransferPayload object."""
    # GIVEN a TransferPayload object with two MiriaObject objects with untrimmed source paths
    for miria_file in transfer_payload.files_to_transfer:
        assert miria_file.source.startswith(ROOT_TO_TRIM)

    # WHEN trimming the source directory
    transfer_payload.trim_paths(attribute_to_trim=SOURCE_ATTRIBUTE)

    # THEN the source directories should no longer contain /home
    for miria_file in transfer_payload.files_to_transfer:
        assert not miria_file.source.startswith(ROOT_TO_TRIM)


def test_transfer_payload_correct_destination_root(transfer_payload: TransferPayload):
    """Tests trimming all destination paths in the TransferPayload object."""

    # GIVEN a TransferPayload object with two MiriaObject objects with untrimmed destination paths
    for miria_file in transfer_payload.files_to_transfer:
        miria_file.destination = ROOT_TO_TRIM + miria_file.destination
        assert miria_file.destination.startswith(ROOT_TO_TRIM)

    # WHEN trimming the destination directories
    transfer_payload.trim_paths(attribute_to_trim=DESTINATION_ATTRIBUTE)

    # THEN the destination directories should no longer contain /home
    for miria_file in transfer_payload.files_to_transfer:
        assert not miria_file.destination.startswith(ROOT_TO_TRIM)


def test_auth_header_old_token(ddn_dataflow_client: DDNDataFlowClient, old_timestamp: datetime):
    """Tests that the refresh method is called if the auth token is too old."""

    # GIVEN a DDNDataFlowClient with an old auth token
    ddn_dataflow_client.token_expiration = old_timestamp

    # WHEN asking for the auth header
    with mock.patch.object(ddn_dataflow_client, "_refresh_auth_token") as mock_refresh_method:
        auth_header: dict = ddn_dataflow_client.auth_header

    # THEN the refresh method should have been called
    mock_refresh_method.assert_called()

    # THEN the returned dict should have the correct format and contain the current auth token
    assert auth_header.get("Authorization") == f"Bearer {ddn_dataflow_client.auth_token}"


def test_auth_header_new_token(ddn_dataflow_client: DDNDataFlowClient):
    """Tests that the refresh method is not called if the auth token is new."""

    # GIVEN a DDNDataFlowClient with a new auth token
    ddn_dataflow_client.token_expiration = datetime.now() + timedelta(minutes=20)

    # WHEN asking for the auth header
    with mock.patch.object(ddn_dataflow_client, "_refresh_auth_token") as mock_refresh_method:
        auth_header: dict = ddn_dataflow_client.auth_header

    # THEN the refresh method should not have been called
    mock_refresh_method.assert_not_called()

    # THEN the returned dict should have the correct format and contain the current auth token
    assert auth_header.get("Authorization") == f"Bearer {ddn_dataflow_client.auth_token}"


def test__refresh_auth_token(ddn_dataflow_client: DDNDataFlowClient, ok_response: Response):
    """Tests if the refresh token is correctly updated when used."""

    # GIVEN a DDNDataFlowClient with new token
    new_token: str = "new_token"
    new_expiration: datetime = datetime.now() + timedelta(minutes=30)
    ok_response._content = WriteStream.write_stream_from_content(
        content={
            "access": new_token,
            "expire": int(new_expiration.timestamp()),
        },
        file_format=FileFormat.JSON,
    ).encode()

    # WHEN refreshing the auth token
    with mock.patch(
        FUNCTION_TO_MOCK,
        return_value=ok_response,
    ):
        ddn_dataflow_client._refresh_auth_token()

    # THEN the api token and the expiration time should be updated
    assert ddn_dataflow_client.auth_token == new_token
    assert ddn_dataflow_client.token_expiration.second == new_expiration.second


def test_archive_file(
    ddn_dataflow_client: DDNDataFlowClient,
    remote_storage_repository: str,
    local_storage_repository: str,
    file_and_sample: FileAndSample,
    trimmed_local_path: str,
    ok_miria_response,
):
    """Tests that the archiving function correctly formats the input and sends API request."""

    # GIVEN two paths that should be archived
    # WHEN running the archive method and providing two paths
    with mock.patch.object(
        APIRequest,
        "api_request_from_content",
        return_value=ok_miria_response,
    ) as mock_request_submitter:
        job_id: int = ddn_dataflow_client.archive_file(file_and_sample)

    # THEN an integer should be returned
    assert isinstance(job_id, int)

    # THEN the mocked submit function should have been called exactly once with correct arguments
    mock_request_submitter.assert_called_once_with(
        api_method=APIMethods.POST,
        url=urljoin(base=ddn_dataflow_client.url, url=DataflowEndpoints.ARCHIVE_FILES),
        headers=dict(ddn_dataflow_client.headers, **ddn_dataflow_client.auth_header),
        json={
            "pathInfo": [
                {
                    "source": local_storage_repository + trimmed_local_path,
                    "destination": remote_storage_repository + file_and_sample.sample.internal_id,
                }
            ],
            "osType": OSTYPE,
            "createFolder": True,
            "metadataList": get_metadata(file_and_sample.sample),
            "settings": [],
        },
        verify=False,
    )


def test_retrieve_files(
    ddn_dataflow_client: DDNDataFlowClient,
    remote_storage_repository: str,
    local_storage_repository: str,
    archive_store: Store,
    trimmed_local_path: str,
    file_and_sample: FileAndSample,
    ok_miria_response,
    retrieve_request_json,
):
    """Tests that the retrieve function correctly formats the input and sends API request."""

    # GIVEN a file and sample which is archived

    # WHEN running retrieve_files and providing a FileAndSample object
    with mock.patch.object(
        APIRequest, "api_request_from_content", return_value=ok_miria_response
    ) as mock_request_submitter:
        job_id: int = ddn_dataflow_client.retrieve_files(files_and_samples=[file_and_sample])

        # THEN an integer should be returned
    assert isinstance(job_id, int)

    # THEN the mocked submit function should have been called exactly once with correct arguments
    retrieve_request_json["pathInfo"][0]["source"] += "/" + Path(file_and_sample.file.path).name
    mock_request_submitter.assert_called_once_with(
        api_method=APIMethods.POST,
        url=urljoin(base=ddn_dataflow_client.url, url=DataflowEndpoints.RETRIEVE_FILES),
        headers=dict(ddn_dataflow_client.headers, **ddn_dataflow_client.auth_header),
        json=retrieve_request_json,
        verify=False,
    )


def test_create_transfer_request_archiving(
    ddn_dataflow_client: DDNDataFlowClient, miria_file_archive: MiriaObject
):
    """Tests creating an archiving request."""
    # GIVEN a TransferData object with an untrimmed source path, and without the source and
    # destination repositories pre-pended
    assert miria_file_archive.source.startswith(ROOT_TO_TRIM)

    # WHEN creating the archiving request
    transfer_request: TransferPayload = ddn_dataflow_client.create_transfer_request(
        [miria_file_archive], is_archiving_request=True
    )

    # THEN the source directories should no longer contain /home
    # THEN the source path should start with the local_storage prefix
    # THEN the destination path should start with the archive prefix
    assert transfer_request.files_to_transfer
    for transfer_data_archive in transfer_request.files_to_transfer:
        assert ROOT_TO_TRIM not in transfer_data_archive.source
        assert transfer_data_archive.source.startswith(ddn_dataflow_client.local_storage)
        assert transfer_data_archive.destination.startswith(ddn_dataflow_client.archive_repository)


def test_create_transfer_request_retrieve(
    ddn_dataflow_client: DDNDataFlowClient, miria_file_retrieve: MiriaObject
):
    """Tests creating a retrieve request."""
    # GIVEN a TransferData object with an untrimmed destination path, and without the source and
    # destination repositories pre-pended
    assert miria_file_retrieve.destination.startswith(ROOT_TO_TRIM)

    # WHEN creating the retrieve request
    transfer_request: TransferPayload = ddn_dataflow_client.create_transfer_request(
        [miria_file_retrieve], is_archiving_request=False
    )

    # THEN the destination directories should no longer contain /home
    # THEN the source path should start with the archive prefix
    # THEN the destination path should start with the local_storage prefix
    assert transfer_request.files_to_transfer
    for transfer_data_archive in transfer_request.files_to_transfer:
        assert ROOT_TO_TRIM not in transfer_data_archive.destination
        assert transfer_data_archive.source.startswith(ddn_dataflow_client.archive_repository)
        assert transfer_data_archive.destination.startswith(ddn_dataflow_client.local_storage)



"""Module to parse the illumina demultiplexing log file."""

from cg.io.json import read_json
from pathlib import Path
from pydantic import BaseModel, Field


class DemuxSoftware(BaseModel):
    software_version: str = Field(..., alias="dragen_version")
    software: str = Field(default="dragen")


class IlluminaDemuxVersionService:
    @staticmethod
    def _parse_demux_version_file(demux_version_file: Path) -> DemuxSoftware:
        """Parse the demultiplexing log file."""
        content = read_json(demux_version_file)
        system_info: dict = content.get("system")
        return DemuxSoftware(**system_info)

    def get_demux_software(self, demux_version_file: Path):
        """Get the demux software."""
        demux_software: DemuxSoftware = self._parse_demux_version_file(demux_version_file)
        return demux_software.software

    def get_demux_software_version(self, demux_version_file: Path) -> str:
        """Get the demultiplexing software version."""
        demux_software: DemuxSoftware = self._parse_demux_version_file(demux_version_file)
        return demux_software.software_version



from pathlib import Path

from cg.constants.constants import ReadDirection, FileFormat, FileExtensions

from cg.services.fastq_concatenation_service.fastq_concatenation_service import (
    FastqConcatenationService,
)
from cg.services.fastq_concatenation_service.utils import generate_concatenated_fastq_delivery_path
from cg.services.deliver_files.delivery_file_fetcher_service.models import SampleFile
from cg.services.deliver_files.delivery_file_formatter_service.models import FormattedFile
from cg.services.deliver_files.delivery_file_formatter_service.utils.sample_file_formatter import (
    SampleFileFormatter,
)


class SampleFileConcatenationFormatter(SampleFileFormatter):
    """
    Format the sample files to deliver, concatenate fastq files and return the formatted files.
    Used for workflows: Microsalt and Mutant.
    """

    def __init__(self, concatenation_service: FastqConcatenationService):
        self.concatenation_service = concatenation_service

    def format_files(
        self, moved_files: list[SampleFile], ticket_dir_path: Path
    ) -> list[FormattedFile]:
        """Format the sample files to deliver, concatenate fastq files and return the formatted files."""
        formatted_files: list[FormattedFile] = super().format_files(
            moved_files=moved_files, ticket_dir_path=ticket_dir_path
        )
        forward_paths, reverse_path = self._concatenate_fastq_files(formatted_files=formatted_files)
        self._replace_fastq_paths(
            reverse_paths=reverse_path,
            forward_paths=forward_paths,
            formatted_files=formatted_files,
        )
        return formatted_files

    def _concatenate_fastq_files(
        self, formatted_files: list[FormattedFile]
    ) -> tuple[list[Path], list[Path]]:
        unique_sample_dir_paths: set[Path] = self._get_unique_sample_paths(
            sample_files=formatted_files
        )
        forward_paths: list[Path] = []
        reverse_paths: list[Path] = []
        for fastq_directory in unique_sample_dir_paths:
            sample_name: str = fastq_directory.name

            forward_path: Path = generate_concatenated_fastq_delivery_path(
                fastq_directory=fastq_directory,
                sample_name=sample_name,
                direction=ReadDirection.FORWARD,
            )
            forward_paths.append(forward_path)
            reverse_path: Path = generate_concatenated_fastq_delivery_path(
                fastq_directory=fastq_directory,
                sample_name=sample_name,
                direction=ReadDirection.REVERSE,
            )
            reverse_paths.append(reverse_path)
            self.concatenation_service.concatenate(
                fastq_directory=fastq_directory,
                forward_output_path=forward_path,
                reverse_output_path=reverse_path,
                remove_raw=True,
            )
        return forward_paths, reverse_paths

    @staticmethod
    def _get_unique_sample_paths(sample_files: list[FormattedFile]) -> set[Path]:
        sample_paths: list[Path] = []
        for sample_file in sample_files:
            sample_paths.append(sample_file.formatted_path.parent)
        return set(sample_paths)

    @staticmethod
    def _replace_fastq_formatted_file_path(
        formatted_files: list[FormattedFile],
        direction: ReadDirection,
        new_path: Path,
    ) -> None:
        """Replace the formatted file path with the new path."""
        for formatted_file in formatted_files:
            if (
                formatted_file.formatted_path.parent == new_path.parent
                and f"{FileFormat.FASTQ}{FileExtensions.GZIP}" in formatted_file.formatted_path.name
                and f"R{direction}" in formatted_file.formatted_path.name
            ):
                formatted_file.formatted_path = new_path

    def _replace_fastq_paths(
        self,
        forward_paths: list[Path],
        reverse_paths: list[Path],
        formatted_files: list[FormattedFile],
    ) -> None:
        """Replace the fastq file paths with the new concatenated fastq file paths."""
        for forward_path in forward_paths:
            self._replace_fastq_formatted_file_path(
                formatted_files=formatted_files,
                direction=ReadDirection.FORWARD,
                new_path=forward_path,
            )
        for reverse_path in reverse_paths:
            self._replace_fastq_formatted_file_path(
                formatted_files=formatted_files,
                direction=ReadDirection.REVERSE,
                new_path=reverse_path,
            )



import fnmatch
from pathlib import Path
from typing import Callable

import pytest

from cg.constants import FileExtensions
from cg.services.illumina.backup.backup_service import IlluminaBackupService
from cg.services.pdc_service.pdc_service import PdcService
from cg.meta.encryption.encryption import EncryptionAPI
from cg.meta.tar.tar import TarAPI
from cg.models.cg_config import PDCArchivingDirectory
from cg.models.cg_config import CGConfig
from cg.store.store import Store


@pytest.fixture
def mock_pdc_query_method(archived_illumina_runs: list[str]) -> Callable:
    """Returns a mock method mimicking the pattern search made by the dsmc q archive command."""

    def mock_method(search_pattern: str) -> list[str] | None:
        if match := fnmatch.filter(archived_illumina_runs, search_pattern):
            return match

    return mock_method


@pytest.fixture
def dsmc_q_archive_output() -> list[str]:
    output: str = """IBM Tivoli Storage Manager
Command Line Backup-Archive Client Interface
  Client Version 7, Release 1, Level 4.0
  Client date/time: 09/22/2023 10:42:42
(c) Copyright by IBM Corporation and other(s) 1990, 2015. All Rights Reserved.

Node Name: HASTA.SCILIFELAB.SE_CLINICAL
Session established with server BLACKHOLE: Linux/x86_64
  Server Version 8, Release 1, Level 9.300
  Server date/time: 09/22/2023 10:42:42  Last access: 09/22/2023 10:42:26

Accessing as node: SLLCLINICAL
             Size  Archive Date - Time    File - Expires on - Description
             ----  -------------------    -------------------------------
           607  B  04/07/2019 07:30:15    /home/hiseq.clinical/ENCRYPT/190329_A00689_0018_AHVKJCDRXX.key.gpg Never Archive Date: 04/07/2019
 1,244,997,334 KB  04/07/2019 04:00:05    /home/hiseq.clinical/ENCRYPT/190329_A00689_0018_AHVKJCDRXX.tar.gz.gpg Never Archive Date: 04/07/2019"""
    return output.splitlines()


@pytest.fixture
def archived_illumina_runs(pdc_archiving_directory: PDCArchivingDirectory) -> list[str]:
    """Returns a list of archived flow cells."""
    return [
        f"{pdc_archiving_directory.current}/new_flow_cell{FileExtensions.TAR}{FileExtensions.GZIP}{FileExtensions.GPG}",
        f"{pdc_archiving_directory.nas}/old_flow_cell{FileExtensions.TAR}{FileExtensions.GZIP}{FileExtensions.GPG}",
        f"{pdc_archiving_directory.pre_nas}/ancient_flow_cell{FileExtensions.TAR}{FileExtensions.GZIP}{FileExtensions.GPG}",
    ]


@pytest.fixture
def backup_api(
    cg_context: CGConfig, illumina_sequencing_runs_directory: Path
) -> IlluminaBackupService:
    """Return a BackupAPI instance."""
    encryption_api: EncryptionAPI = EncryptionAPI(
        binary_path=cg_context.encryption.binary_path, dry_run=True
    )
    store: Store = cg_context.status_db
    tar_api: TarAPI = TarAPI(binary_path=cg_context.tar.binary_path, dry_run=True)
    pdc_service: PdcService = PdcService(binary_path=cg_context.pdc.binary_path, dry_run=True)

    pdc_archiving_directory: PDCArchivingDirectory = (
        cg_context.illumina_backup_service.pdc_archiving_directory
    )
    _backup_api: IlluminaBackupService = IlluminaBackupService(
        encryption_api=encryption_api,
        status_db=store,
        tar_api=tar_api,
        pdc_service=pdc_service,
        pdc_archiving_directory=pdc_archiving_directory,
        sequencing_runs_dir=illumina_sequencing_runs_directory,
        dry_run=True,
    )
    return _backup_api


@pytest.fixture
def spring_file_path() -> Path:
    """Return spring file path"""
    return Path("/path/to/spring_file.spring")


@pytest.fixture
def backup_file_path() -> str:
    """Return path to a file used in the backup process"""
    return "/path/to/backup_file.extension"


@pytest.fixture
def archived_illumina_run() -> Path:
    """Path of archived sequencing run"""
    return Path("/path/to/archived/run_devices.tar.gz.gpg")


@pytest.fixture
def archived_key() -> Path:
    """Path of archived key"""
    return Path("/path/to/archived/encryption_key.key.gpg")



import shutil
from pathlib import Path
from cg.constants.delivery import INBOX_NAME
from cg.services.deliver_files.delivery_file_fetcher_service.models import (
    DeliveryFiles,
    DeliveryMetaData,
    CaseFile,
    SampleFile,
)
from cg.utils.files import link_or_overwrite_file


class DeliveryFilesMover:
    """
    Class that encapsulates the logic required for moving files to the customer folder.
    """

    def move_files(self, delivery_files: DeliveryFiles, delivery_base_path: Path) -> DeliveryFiles:
        """Move the files to the customer folder."""
        inbox_dir_path: Path = self._create_ticket_inbox_dir_path(
            delivery_base_path=delivery_base_path, delivery_data=delivery_files.delivery_data
        )
        self._create_ticket_inbox_folder(inbox_dir_path)
        self._create_hard_links_for_delivery_files(
            delivery_files=delivery_files, inbox_dir_path=inbox_dir_path
        )
        return self._replace_file_paths_with_inbox_dir_paths(
            delivery_files=delivery_files, inbox_dir_path=inbox_dir_path
        )

    @staticmethod
    def _create_ticket_inbox_folder(
        inbox_dir_path: Path,
    ) -> Path:
        """Create a ticket inbox folder in the customer folder, overwrites if already present."""
        inbox_dir_path.mkdir(parents=True, exist_ok=True)
        return inbox_dir_path

    @staticmethod
    def _create_ticket_inbox_dir_path(
        delivery_base_path: Path, delivery_data: DeliveryMetaData
    ) -> Path:
        """Create the path to the ticket inbox folder."""
        return Path(
            delivery_base_path,
            delivery_data.customer_internal_id,
            INBOX_NAME,
            delivery_data.ticket_id,
        )

    @staticmethod
    def _create_inbox_file_path(file_path: Path, inbox_dir_path: Path) -> Path:
        """Create the path to the inbox file."""
        return Path(inbox_dir_path, file_path.name)

    def _create_hard_link_file_paths(
        self, file_models: list[SampleFile | CaseFile], inbox_dir_path: Path
    ) -> None:
        """Create hard links to the sample files in the customer folder."""
        for file_model in file_models:
            inbox_file_path: Path = self._create_inbox_file_path(
                file_path=file_model.file_path, inbox_dir_path=inbox_dir_path
            )
            link_or_overwrite_file(src=file_model.file_path, dst=inbox_file_path)

    def _create_hard_links_for_delivery_files(
        self, delivery_files: DeliveryFiles, inbox_dir_path: Path
    ) -> None:
        """Create hard links to the files in the customer folder."""
        if delivery_files.case_files:
            self._create_hard_link_file_paths(
                file_models=delivery_files.case_files, inbox_dir_path=inbox_dir_path
            )
        self._create_hard_link_file_paths(
            file_models=delivery_files.sample_files, inbox_dir_path=inbox_dir_path
        )

    def _replace_file_path_with_inbox_dir_path(
        self, file_models: list[SampleFile | CaseFile], inbox_dir_path: Path
    ) -> list[SampleFile | CaseFile]:
        """Replace the file path with the inbox path."""
        for file_model in file_models:
            inbox_file_path: Path = self._create_inbox_file_path(
                file_path=file_model.file_path, inbox_dir_path=inbox_dir_path
            )
            file_model.file_path = inbox_file_path
        return file_models

    def _replace_file_paths_with_inbox_dir_paths(
        self,
        delivery_files: DeliveryFiles,
        inbox_dir_path: Path,
    ) -> DeliveryFiles:
        """
        Replace to original file paths in the delivery files with the customer inbox file paths.
        """
        if delivery_files.case_files:
            delivery_files.case_files = self._replace_file_path_with_inbox_dir_path(
                file_models=delivery_files.case_files, inbox_dir_path=inbox_dir_path
            )
        delivery_files.sample_files = self._replace_file_path_with_inbox_dir_path(
            file_models=delivery_files.sample_files, inbox_dir_path=inbox_dir_path
        )
        return delivery_files



"""Fixtures for PacBio metrics objects."""

import pytest

from cg.constants.pacbio import (
    CCSAttributeIDs,
    ControlAttributeIDs,
    LoadingAttributesIDs,
    PolymeraseDataAttributeIDs,
    SmrtLinkDatabasesIDs,
)
from cg.services.run_devices.pacbio.metrics_parser.models import (
    ControlMetrics,
    PacBioMetrics,
    PolymeraseMetrics,
    ProductivityMetrics,
    ReadMetrics,
    SmrtlinkDatasetsMetrics,
)


@pytest.fixture
def pac_bio_read_metrics() -> ReadMetrics:
    data: dict[str, any] = {
        CCSAttributeIDs.HIFI_READS: 6580977,
        CCSAttributeIDs.HIFI_YIELD: 106275091861,
        CCSAttributeIDs.HIFI_MEAN_READ_LENGTH: 16148,
        CCSAttributeIDs.HIFI_MEDIAN_READ_LENGTH: 16218,
        CCSAttributeIDs.HIFI_READ_LENGTH_N50: 16218,
        CCSAttributeIDs.HIFI_MEDIAN_READ_QUALITY: 34,
        CCSAttributeIDs.PERCENT_Q30: 93.2,
        CCSAttributeIDs.FAILED_READS: 281284,
        CCSAttributeIDs.FAILED_YIELD: 4862922407,
        CCSAttributeIDs.FAILED_MEAN_READ_LENGTH: 17288,
    }
    return ReadMetrics.model_validate(data, from_attributes=True)


@pytest.fixture
def pac_bio_control_metrics() -> ControlMetrics:
    data: dict[str, any] = {
        ControlAttributeIDs.NUMBER_OF_READS: 2750,
        ControlAttributeIDs.MEAN_READ_LENGTH: 57730,
        ControlAttributeIDs.PERCENT_MEAN_READ_CONCORDANCE: 0.906334,
        ControlAttributeIDs.PERCENT_MODE_READ_CONCORDANCE: 0.91,
    }
    return ControlMetrics.model_validate(data, from_attributes=True)


@pytest.fixture
def pac_bio_productivity_metrics() -> ProductivityMetrics:
    data: dict[str, any] = {
        LoadingAttributesIDs.PRODUCTIVE_ZMWS: 25165824,
        LoadingAttributesIDs.P_0: 10012557,
        LoadingAttributesIDs.P_1: 15048838,
        LoadingAttributesIDs.P_2: 104429,
    }
    return ProductivityMetrics.model_validate(data, from_attributes=True)


@pytest.fixture
def pac_bio_polymerase_metrics() -> PolymeraseMetrics:
    data: dict[str, any] = {
        PolymeraseDataAttributeIDs.MEAN_READ_LENGTH: 82182,
        PolymeraseDataAttributeIDs.READ_LENGTH_N50: 153250,
        PolymeraseDataAttributeIDs.MEAN_LONGEST_SUBREAD_LENGTH: 18466,
        PolymeraseDataAttributeIDs.LONGEST_SUBREAD_LENGTH_N50: 22250,
    }
    return PolymeraseMetrics.model_validate(data, from_attributes=True)


@pytest.fixture
def pac_bio_smrtlink_databases_metrics(
    smrt_cell_internal_id: str, pac_bio_sample_internal_id: str, pac_bio_1_a01_cell_full_name: str
) -> SmrtlinkDatasetsMetrics:
    data: dict[str, any] = {
        SmrtLinkDatabasesIDs.BIO_SAMPLE_NAME: pac_bio_sample_internal_id,
        SmrtLinkDatabasesIDs.CELL_ID: smrt_cell_internal_id,
        SmrtLinkDatabasesIDs.CELL_INDEX: 0,
        SmrtLinkDatabasesIDs.MOVIE_NAME: pac_bio_1_a01_cell_full_name,
        SmrtLinkDatabasesIDs.PATH: "/srv/cg_data/pacbio/r84202_20240522_133539/1_A01/pb_formats/m84202_240522_135641_s1.hifi_reads.consensusreadset.xml",
        SmrtLinkDatabasesIDs.RUN_COMPLETED_AT: "2024-05-24T02:21:20.970Z",
        SmrtLinkDatabasesIDs.WELL_NAME: "A01",
        SmrtLinkDatabasesIDs.WELL_SAMPLE_NAME: pac_bio_sample_internal_id,
    }
    return SmrtlinkDatasetsMetrics.model_validate(data, from_attributes=True)


@pytest.fixture
def pac_bio_metrics(
    pac_bio_read_metrics: ReadMetrics,
    pac_bio_control_metrics: ControlMetrics,
    pac_bio_productivity_metrics: ProductivityMetrics,
    pac_bio_polymerase_metrics: PolymeraseMetrics,
    pac_bio_smrtlink_databases_metrics: SmrtlinkDatasetsMetrics,
) -> PacBioMetrics:
    return PacBioMetrics(
        read=pac_bio_read_metrics,
        control=pac_bio_control_metrics,
        productivity=pac_bio_productivity_metrics,
        polymerase=pac_bio_polymerase_metrics,
        dataset_metrics=pac_bio_smrtlink_databases_metrics,
    )



"""Module that holds the error handlers for the Post Processing service."""

import logging

from cg.exc import CgError

LOG = logging.getLogger(__name__)


def handle_post_processing_errors(to_except: tuple, to_raise):
    """Handle errors for the Post processing services."""

    def error_handling(fn):
        def wrapper(*args, **kwargs):
            try:
                return fn(*args, **kwargs)
            except to_except as error:
                raise to_raise(error) from error
            except Exception as error:
                raise CgError(f"{error}") from error

        return wrapper

    return error_handling



import json
import logging
import tempfile
from http import HTTPStatus
from pathlib import Path

from flask import Blueprint, abort, g, jsonify, make_response, request
from pydantic.v1 import ValidationError
from requests.exceptions import HTTPError
from sqlalchemy.exc import IntegrityError
from urllib3.exceptions import MaxRetryError, NewConnectionError
from werkzeug.utils import secure_filename

from cg.apps.orderform.excel_orderform_parser import ExcelOrderformParser
from cg.apps.orderform.json_orderform_parser import JsonOrderformParser
from cg.constants import ANALYSIS_SOURCES, METAGENOME_SOURCES
from cg.constants.constants import FileFormat
from cg.exc import (
    OrderError,
    OrderExistsError,
    OrderFormError,
    OrderNotDeliverableError,
    OrderNotFoundError,
    TicketCreationError,
)
from cg.io.controller import WriteStream
from cg.meta.orders import OrdersAPI
from cg.meta.orders.ticket_handler import TicketHandler
from cg.models.orders.order import OrderIn, OrderType
from cg.models.orders.orderform_schema import Orderform
from cg.server.dto.delivery_message.delivery_message_response import (
    DeliveryMessageResponse,
)
from cg.server.dto.orders.order_delivery_update_request import (
    OrderDeliveredUpdateRequest,
)
from cg.server.dto.orders.order_patch_request import OrderDeliveredPatch
from cg.server.dto.orders.orders_request import OrdersRequest
from cg.server.dto.orders.orders_response import Order, OrdersResponse
from cg.server.endpoints.utils import before_request
from cg.server.ext import (
    db,
    delivery_message_service,
    lims,
    order_service,
    order_submitter_registry,
    ticket_handler,
)
from cg.store.models import Application, Customer

ORDERS_BLUEPRINT = Blueprint("orders", __name__, url_prefix="/api/v1")
ORDERS_BLUEPRINT.before_request(before_request)

LOG = logging.getLogger(__name__)


@ORDERS_BLUEPRINT.route("/orders")
def get_orders():
    """Return the latest orders."""
    data = OrdersRequest.model_validate(request.args.to_dict())
    response: OrdersResponse = order_service.get_orders(data)
    return make_response(response.model_dump())


@ORDERS_BLUEPRINT.route("/orders/<order_id>")
def get_order(order_id: int):
    """Return an order."""
    try:
        response: Order = order_service.get_order(order_id)
        response_dict: dict = response.model_dump()
        return make_response(response_dict)
    except OrderNotFoundError as error:
        return make_response(jsonify(error=str(error)), HTTPStatus.NOT_FOUND)


@ORDERS_BLUEPRINT.route("/orders/<order_id>/delivered", methods=["PATCH"])
def set_order_delivered(order_id: int):
    try:
        request_data = OrderDeliveredPatch.model_validate(request.json)
        delivered: bool = request_data.delivered
        response_data: Order = order_service.set_delivery(order_id=order_id, delivered=delivered)
        return jsonify(response_data.model_dump()), HTTPStatus.OK
    except OrderNotFoundError as error:
        return jsonify(error=str(error)), HTTPStatus.NOT_FOUND


@ORDERS_BLUEPRINT.route("/orders/<order_id>/update-delivery-status", methods=["POST"])
def update_order_delivered(order_id: int):
    """Update the delivery status of an order based on the number of delivered analyses."""
    try:
        request_data = OrderDeliveredUpdateRequest.model_validate(request.json)
        delivered_analyses: int = request_data.delivered_analyses_count
        order_service.update_delivered(order_id=order_id, delivered_analyses=delivered_analyses)
    except OrderNotFoundError as error:
        return jsonify(error=str(error)), HTTPStatus.NOT_FOUND


@ORDERS_BLUEPRINT.route("/orders/<order_id>/delivery_message")
def get_delivery_message_for_order(order_id: int):
    """Return the delivery message for an order."""
    try:
        response: DeliveryMessageResponse = delivery_message_service.get_order_message(order_id)
        response_dict: dict = response.model_dump()
        return make_response(response_dict)
    except OrderNotDeliverableError as error:
        return make_response(jsonify(error=str(error)), HTTPStatus.PRECONDITION_FAILED)
    except OrderNotFoundError as error:
        return make_response(jsonify(error=str(error))), HTTPStatus.NOT_FOUND


@ORDERS_BLUEPRINT.route("/orderform", methods=["POST"])
def create_order_from_form():
    """Parse an orderform/JSON export."""
    input_file = request.files.get("file")
    filename = secure_filename(input_file.filename)

    error_message: str
    try:
        if filename.lower().endswith(".xlsx"):
            temp_dir = Path(tempfile.gettempdir())
            saved_path = str(temp_dir / filename)
            input_file.save(saved_path)
            order_parser = ExcelOrderformParser()
            order_parser.parse_orderform(excel_path=saved_path)
        else:
            json_data = json.load(input_file.stream, strict=False)
            order_parser = JsonOrderformParser()
            order_parser.parse_orderform(order_data=json_data)
        parsed_order: Orderform = order_parser.generate_orderform()
    except (  # user misbehaviour
        AttributeError,
        OrderFormError,
        OverflowError,
        ValidationError,
        ValueError,
    ) as error:
        error_message = error.message if hasattr(error, "message") else str(error)
        LOG.error(error_message)
        http_error_response = HTTPStatus.BAD_REQUEST
    except (  # system misbehaviour
        NewConnectionError,
        MaxRetryError,
        TimeoutError,
        TypeError,
    ) as error:
        LOG.exception(error)
        error_message = error.message if hasattr(error, "message") else str(error)
        http_error_response = HTTPStatus.INTERNAL_SERVER_ERROR
    else:
        return jsonify(**parsed_order.model_dump())

    if error_message:
        return abort(make_response(jsonify(message=error_message), http_error_response))


@ORDERS_BLUEPRINT.route("/submit_order/<order_type>", methods=["POST"])
def submit_order(order_type):
    """Submit an order for samples."""
    api = OrdersAPI(
        lims=lims,
        status=db,
        ticket_handler=ticket_handler,
        submitter_registry=order_submitter_registry,
    )
    error_message: str
    try:
        request_json = request.get_json()
        LOG.info(
            "processing order: %s",
            WriteStream.write_stream_from_content(
                content=request_json, file_format=FileFormat.JSON
            ),
        )
        project = OrderType(order_type)
        order_in = OrderIn.parse_obj(request_json, project=project)
        existing_ticket: str | None = ticket_handler.parse_ticket_number(order_in.name)
        if existing_ticket and order_service.store.get_order_by_ticket_id(existing_ticket):
            raise OrderExistsError(f"Order with ticket id {existing_ticket} already exists.")

        result: dict = api.submit(
            project=project,
            order_in=order_in,
            user_name=g.current_user.name,
            user_mail=g.current_user.email,
        )
        order_service.create_order(order_in)

    except (  # user misbehaviour
        OrderError,
        OrderExistsError,
        OrderFormError,
        ValidationError,
        ValueError,
    ) as error:
        error_message = error.message if hasattr(error, "message") else str(error)
        http_error_response = HTTPStatus.BAD_REQUEST
        LOG.error(error_message)
    except (  # system misbehaviour
        AttributeError,
        ConnectionError,
        HTTPError,
        IntegrityError,
        KeyError,
        NewConnectionError,
        MaxRetryError,
        TimeoutError,
        TicketCreationError,
        TypeError,
    ) as error:
        LOG.exception(error)
        error_message = error.message if hasattr(error, "message") else str(error)
        http_error_response = HTTPStatus.INTERNAL_SERVER_ERROR
    else:
        return jsonify(
            project=result["project"], records=[record.to_dict() for record in result["records"]]
        )

    if error_message:
        return abort(make_response(jsonify(message=error_message), http_error_response))


@ORDERS_BLUEPRINT.route("/options")
def get_options():
    """Return various options."""
    customers: list[Customer | None] = (
        db.get_customers() if g.current_user.is_admin else g.current_user.customers
    )

    app_tag_groups: dict[str, list[str]] = {"ext": []}
    applications: list[Application] = db.get_applications_is_not_archived()
    for application in applications:
        if not application.versions:
            LOG.debug(f"Skipping application {application} that doesn't have a price")
            continue
        if application.is_external:
            app_tag_groups["ext"].append(application.tag)
        if application.prep_category not in app_tag_groups:
            app_tag_groups[application.prep_category]: list[str] = []
        app_tag_groups[application.prep_category].append(application.tag)

    source_groups = {"metagenome": METAGENOME_SOURCES, "analysis": ANALYSIS_SOURCES}

    return jsonify(
        applications=app_tag_groups,
        beds=[bed.name for bed in db.get_active_beds()],
        customers=[
            {
                "text": f"{customer.name} ({customer.internal_id})",
                "value": customer.internal_id,
                "isTrusted": customer.is_trusted,
            }
            for customer in customers
        ],
        organisms=[
            {
                "name": organism.name,
                "reference_genome": organism.reference_genome,
                "internal_id": organism.internal_id,
                "verified": organism.verified,
            }
            for organism in db.get_all_organisms()
        ],
        panels=[panel.abbrev for panel in db.get_panels()],
        sources=source_groups,
    )



import logging
from datetime import datetime
from pathlib import Path

from cg.constants import Workflow
from cg.services.deliver_files.deliver_files_service.deliver_files_service import (
    DeliverFilesService,
)
from cg.services.deliver_files.deliver_files_service.deliver_files_service_factory import (
    DeliveryServiceFactory,
)
from cg.store.models import Case, Analysis
from cg.store.store import Store


LOG = logging.getLogger(__name__)


def deliver_raw_data_for_analyses(
    analyses: list[Analysis],
    status_db: Store,
    delivery_path: Path,
    service_builder: DeliveryServiceFactory,
    dry_run: bool,
):
    """Deliver raw data for a list of analyses"""
    for analysis in analyses:
        try:
            case: Case = analysis.case
            delivery_service: DeliverFilesService = service_builder.build_delivery_service(
                delivery_type=case.data_delivery,
                workflow=Workflow.FASTQ,
            )

            delivery_service.deliver_files_for_case(
                case=case, delivery_base_path=delivery_path, dry_run=dry_run
            )
            status_db.update_analysis_upload_started_at(
                analysis_id=analysis.id, upload_started_at=datetime.now()
            )
        except Exception as error:
            status_db.update_analysis_upload_started_at(
                analysis_id=analysis.id, upload_started_at=None
            )
            LOG.error(f"Could not deliver files for analysis {analysis.id}: {error}")
            continue



"""Chanjo2 API fixtures."""

from typing import Any
from unittest.mock import Mock

import pytest
import requests

from cg.clients.chanjo2.client import Chanjo2APIClient


@pytest.fixture
def chanjo2_api_client(context_config: dict[str, Any]) -> Chanjo2APIClient:
    return Chanjo2APIClient(context_config["chanjo2"]["host"])


@pytest.fixture
def coverage_post_response_json(sample_id: str) -> dict[str, dict]:
    return {
        sample_id: {
            "mean_coverage": 55.55,
            "coverage_completeness_percent": 33.33,
        }
    }


@pytest.fixture
def coverage_post_response_invalid_values_json(sample_id: str) -> dict[str, dict]:
    return {
        sample_id: {
            "mean_coverage": "I am a string",
            "coverage_completeness_percent": 33.33,
        }
    }


@pytest.fixture
def coverage_post_response_invalid_attributes_json(sample_id: str) -> dict[str, dict]:
    return {
        sample_id: {
            "not_a_metric": 55.55,
            "coverage_completeness_percent": 33.33,
        },
    }


@pytest.fixture
def coverage_post_response_success(coverage_post_response_json: dict) -> Mock:
    post_response = Mock()
    post_response.json.return_value = coverage_post_response_json
    return post_response


@pytest.fixture
def coverage_post_response_http_error() -> Mock:
    post_response = Mock()
    post_response.raise_for_status.side_effect = requests.HTTPError()
    return post_response


@pytest.fixture
def coverage_post_response_invalid_values(coverage_post_response_invalid_values_json: dict) -> Mock:
    post_response = Mock()
    post_response.json.return_value = coverage_post_response_invalid_values_json
    return post_response


@pytest.fixture
def coverage_post_response_invalid_attributes(
    coverage_post_response_invalid_attributes_json: dict,
) -> Mock:
    post_response = Mock()
    post_response.json.return_value = coverage_post_response_invalid_attributes_json
    return post_response


@pytest.fixture
def coverage_post_response_empty() -> Mock:
    post_response = Mock()
    post_response.json.return_value = {}
    return post_response



import logging
from pathlib import Path
from cg.meta.workflow.mutant.quality_controller.models import (
    CaseQualityResult,
    SampleQualityResults,
    SamplesQualityResults,
)

LOG = logging.getLogger(__name__)


def log_results(
    case_quality_result: CaseQualityResult,
    samples_quality_results: SamplesQualityResults,
    report_file_path: Path,
) -> None:
    if case_quality_result.passes_qc:
        case_message = f"QC passed, see {report_file_path} for details."
    else:
        case_message = get_case_fail_message(case_quality_result)
    LOG.warning(case_message)

    samples_message = get_samples_results_message(samples_quality_results)
    LOG.info(samples_message)


def log_sample_result(
    result: SampleQualityResults,
    is_external_negative_control: bool = False,
    is_internal_negative_control: bool = False,
) -> None:
    control_message = ""
    if is_external_negative_control:
        control_message = "External negative control sample "
    if is_internal_negative_control:
        control_message = "Internal negative control sample "
    if result.passes_qc:
        message = f"{control_message}{result.sample_id} passed QC."
        LOG.info(message)
    else:
        message = f"{control_message}{result.sample_id} failed QC."
        LOG.warning(message)


def log_case_result(result: CaseQualityResult) -> None:
    if not result.passes_qc:
        LOG.warning("Case failed QC.")
    else:
        LOG.warning("Case passed QC.")


def get_case_fail_message(case_quality_result: CaseQualityResult) -> str:
    fail_reasons = []
    if not case_quality_result.internal_negative_control_passes_qc:
        fail_reasons.append("The internal negative control sample failed QC.\n")
    if not case_quality_result.external_negative_control_passes_qc:
        fail_reasons.append("The external negative control sample failed QC.\n")

    fail_message = "QC failed."

    return fail_message + "\n".join(fail_reasons)


def get_samples_results_message(samples_quality_results: SamplesQualityResults) -> str:
    internal_negative_control_message: str = "Internal negative control sample " + (
        "passed QC."
        if samples_quality_results.internal_negative_control.passes_qc
        else "failed QC."
    )
    external_negative_control_message: str = "External negative control sample " + (
        "passed QC."
        if samples_quality_results.external_negative_control.passes_qc
        else "failed QC."
    )

    samples_message: str = (
        f"Sample results: {samples_quality_results.total_samples_count} total, {samples_quality_results.failed_samples_count} failed, {samples_quality_results.passed_samples_count} passed."
    )

    return " ".join(
        [internal_negative_control_message, external_negative_control_message, samples_message]
    )



import pytest

from cg.services.deliver_files.delivery_file_fetcher_service.delivery_file_fetcher_service import (
    FetchDeliveryFilesService,
)
from cg.services.deliver_files.delivery_file_fetcher_service.models import DeliveryFiles


@pytest.mark.parametrize(
    "expected_delivery_files,delivery_file_service",
    [
        ("expected_fastq_delivery_files", "fastq_delivery_service"),
        ("expected_analysis_delivery_files", "analysis_delivery_service"),
    ],
)
def test_get_files_to_deliver(
    expected_delivery_files: DeliveryFiles,
    delivery_file_service: FetchDeliveryFilesService,
    case_id: str,
    request,
):
    """Test to get the files to deliver from the FetchFastqDeliverFilesService."""
    # GIVEN a case id, samples that are present in housekeeper and a delivery service
    delivery_file_service = request.getfixturevalue(delivery_file_service)
    expected_delivery_files = request.getfixturevalue(expected_delivery_files)

    # WHEN getting the files to deliver
    delivery_files: DeliveryFiles = delivery_file_service.get_files_to_deliver(case_id)

    # THEN assert that the files to deliver are fetched
    assert delivery_files == expected_delivery_files
